//
//  FlatGLWidget.cpp
//  ParchmentFlattener
//
//  Copyright (c) 2014 Kazim Pal.
//  Released under the MIT License.
//
#ifndef IGL_HEADER_ONLY
#  define IGL_HEADER_ONLY
#endif

#include "FlatGLWidget.h"
#ifdef __APPLE__
#include <GL/glu.h>
#else
#include <GL/glu.h>
#endif
#include <QWheelEvent>
#include <qmath.h>
#include <QGLPixelBuffer>
#include "IGLMesh.h"
#include "ParchmentFlattener.h"
#include <igl/axis_angle_to_quat.h>
#include <igl/quat_to_mat.h>
#include "AutomationEngine.h"
#include <Eigen/Dense>
#include <QTimer>
#include <opencv/highgui.h>
#include <QInputDialog>
#include <QMessageBox>
#include <QVector2D>
#include <QMatrix4x4>

#define Z_NEAR_MAIN 0.001
#define Z_NEAR_PROJ 0.01
#define CLICK_THRESH 10
#define DELTA_ZOOM 0.1
#define MIN_ZOOM 0.1
#define DEFAULT_SUGGESTION_TAIL_LENGTH 10
#define DO_SUGGESTION_TAILS false
#define CLICK_THRESH 10
#define SPIN_AROUND false

bool loadEnhanced = false;
bool doGamma = false;

void perspectiveGL( GLdouble fovY, GLdouble aspect, GLdouble zNear, GLdouble zFar )
{
    const GLdouble pi = 3.1415926535897932384626433832795;
    double fW, fH;

    fH = tan( fovY / 360 * pi ) * zNear;
    fW = fH * aspect;
    qDebug() << fH << ", " << fW;
    glFrustum( -fW, fW, -fH, fH, zNear, zFar );
}

void DrawCircle(double cx, double cy, double cz, double r, int num_segments)
{
    float theta = 2 * 3.1415926 / float(num_segments);
    float tangetial_factor = tanf(theta);
    float radial_factor = cosf(theta);

    float x = r;
    float y = 0;

    glBegin(GL_LINE_LOOP);
    for(int ii = 0; ii < num_segments; ii++)
    {
        glVertex3d(x + cx, y + cy, cz);
        float tx = -y;
        float ty = x;

        x += tx * tangetial_factor;
        y += ty * tangetial_factor;

        x *= radial_factor;
        y *= radial_factor;
    }
    glEnd();
}

void DrawElipse(Vector3d C, Vector3d xaxis, Vector3d yaxis, double rx, double ry, int num_segments)
{
    double dtheta = 2 * 3.1415926 / float(num_segments);

    glBegin(GL_LINE_LOOP);
    for(int ii = 0; ii < num_segments; ii++)
    {
        double theta = ii * dtheta;
        Vector3d dx = rx * qCos(theta) * xaxis;
        Vector3d dy = ry * qSin(theta) * yaxis;
        Vector3d p = C + dx + dy;

        glVertex3d(p(0), p(1), p(2));
    }
    glEnd();
}

FlatGLWidget::FlatGLWidget(void)
{
    init();
}

FlatGLWidget::FlatGLWidget(QWidget *parent, QGLWidget *share) : QGLWidget(parent,share)
{
    init();
}

FlatGLWidget::FlatGLWidget(IGLMesh *_mesh, QWidget *parent, QGLWidget *share) : QGLWidget(parent,share)
{
    mesh = _mesh;
    init();
}

FlatGLWidget::~FlatGLWidget(void)
{
    for(int k = 0; k < numTextures; ++k)
    {
        deleteTexture(meshTextures[k]);
    }
    delete[] meshTextures;
    delete[] quat_rot_current;
    delete[] quat_rot_new;
    delete[] rotation_matrix;
    delete mesh;
    delete automationEngine;
}

void FlatGLWidget::init()
{
    camz = -5;
    currentGravity = Vector3d(0,0,1);
    currentUpVec = Vector3d(0,1,0);
    currentRightVec = currentUpVec.cross(currentGravity);
    znear = Z_NEAR_MAIN;
    zfar = 10;
    setMouseTracking(false);
    quat_rot_current = new double[4];
    quat_rot_new = new double[4];
    rotation_matrix = new double[16];
    resetRotation();
    target = Vector3d(0,0,0);
    target_new = Vector3d(0,0,0);
    renderOriginal = false;

    automationEngine = new AutomationEngine();

    renderConstraints = true;
    renderScaleField = false;
    renderActualMesh = true;
    renderInterpolation = false;
    renderNormals = false;
    interpTimer = 0;
    camz_interp = 1;
    renderEnhanced = false;
    renderStretch = false;

    texturesLoaded = false;
}

void FlatGLWidget::loadTextures()
{
    if(texturesLoaded)
    {
        for(int k = 0; k < numTextures; ++k)
        {
            deleteTexture(meshTextures[k]);
        }
        delete[] meshTextures;
        numTextures = 0;
    }

    QMessageBox::StandardButton reply;
    bool enhancedAvailable = false;
    if(enhancedAvailable)
    {
        reply = QMessageBox::question(this, "", "Shading correction?", QMessageBox::Yes|QMessageBox::No);
        if (reply == QMessageBox::Yes)
        {
            loadEnhanced = true;
        }
    }

    if(!loadEnhanced)
    {
        reply = QMessageBox::question(this, "", "Gamma correction?", QMessageBox::Yes|QMessageBox::No);
        if (reply == QMessageBox::Yes)
        {
            doGamma = true;
        }
    }

    makeCurrent();
    // Load the textures
    glEnable(GL_TEXTURE_2D);
    numTextures = mesh->texFNames.size();
    if(loadEnhanced)
        meshTextures = new GLuint[numTextures*2];
    else
        meshTextures = new GLuint[numTextures];

    for(int k = 0; k < numTextures; ++k)
    {
        //QString texfname = mesh->texFNames.at(k);
        QString texfname = ParchmentFlattener::meshDirectory;
        texfname.append(mesh->texFNames.at(k));
        qDebug() << "Loading texture " << texfname << "\n";
        QImage foo(texfname);
        if(doGamma)
        {
            for(int i = 0; i < foo.width(); ++i)
            {
                for(int j = 0; j < foo.height(); ++j)
                {
                    QRgb rgb_ij = foo.pixel(i,j);
                    QColor col_ij(rgb_ij);
                    double r_ij, g_ij, b_ij;
                    col_ij.getRgbF(&r_ij, &g_ij, &b_ij);
                    double gammapower = 1.0/2.2;
                    r_ij = qPow(r_ij, gammapower);
                    g_ij = qPow(g_ij, gammapower);
                    b_ij = qPow(b_ij, gammapower);
                    col_ij.setRgbF(r_ij,g_ij,b_ij);
                    rgb_ij = col_ij.rgb();
                    foo.setPixel(i,j,rgb_ij);
                }
            }
        }
        meshTextures[k] = bindTexture(foo, GL_TEXTURE_2D, GL_RGBA, QGLContext::LinearFilteringBindOption | QGLContext::InvertedYBindOption);
    }

    if(loadEnhanced)
    {
        for(int k = 0; k < numTextures; ++k)
        {
            QString texfname = ParchmentFlattener::meshDirectory;
            texfname.append(mesh->texFNames.at(k));
            texfname.chop(4);
            texfname.append("E.png");
            qDebug() << "Loading texture " << texfname << "\n";
            QImage foo(texfname);
            if(doGamma)
            {
                for(int i = 0; i < foo.width(); ++i)
                {
                    for(int j = 0; j < foo.height(); ++j)
                    {
                        QRgb rgb_ij = foo.pixel(i,j);
                        QColor col_ij(rgb_ij);
                        double r_ij, g_ij, b_ij;
                        col_ij.getRgbF(&r_ij, &g_ij, &b_ij);
                        double gammapower = 1.0/2.2;
                        r_ij = qPow(r_ij, gammapower);
                        g_ij = qPow(g_ij, gammapower);
                        b_ij = qPow(b_ij, gammapower);
                        col_ij.setRgbF(r_ij,g_ij,b_ij);
                        rgb_ij = col_ij.rgb();
                        foo.setPixel(i,j,rgb_ij);
                    }
                }
            }
            meshTextures[k+numTextures] = bindTexture(foo, GL_TEXTURE_2D, GL_RGBA, QGLContext::LinearFilteringBindOption | QGLContext::InvertedYBindOption);
        }
    }
    texturesLoaded = true;
}

void FlatGLWidget::resetRotation()
{
    return;
    rot_angle_old = 0;
    rot_angle_new = 0;
    for(int i = 0; i < 16; ++i)
    {
        if(i == 0 || i == 5 || i == 10 || i == 15)
            rotation_matrix[i] = 1;
        else
            rotation_matrix[i] = 0;
    }
}

void FlatGLWidget::initializeGL()
{
    qglClearColor(qtBlack);

    glEnable(GL_DEPTH_TEST);
    glDisable(GL_CULL_FACE);
	glEnable(GL_TEXTURE_2D);
    glShadeModel(GL_SMOOTH);
    glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
    setProjectionMatrixGL();
    setModelViewMatrixGL();
}

void FlatGLWidget::paintGL()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    setProjectionMatrixGL();

    setModelViewMatrixGL();
    if(renderInterpolation)
    {
        renderInterpolatedMesh();
    }
    else if(renderOriginal)
    {
        renderOriginalMesh();
        if(renderConstraints || renderScaleField)
        {
            renderLineConstraintsW();
            renderScaleConstraintsW();
        }
    }
    else
    {
        if(mesh->flattened)
        {
            renderMesh();
            if(renderConstraints)
            {
                renderLineConstraints();
                renderScaleConstraints();
            }
        }
    }
}

void FlatGLWidget::resizeGL(int width, int height)
{
    window_width = width;
    window_height = height;
    glViewport(0,0,width,height);
    setProjectionMatrixGL();
}

void FlatGLWidget::mousePressEvent(QMouseEvent * e)
{
    m_lastPoint = e->pos();
    m_mouseClick = true;

    if(ParchmentFlattener::markupMode == ParchmentFlattener::EditLine)
    {
        mesh->selectedLineConstraintPoint = selectLineConstraintKeypoint(e->x(),e->y());
        updateGL();
        //wGlWidget->updateGL();
    }
}

void FlatGLWidget::mouseMoveEvent(QMouseEvent * e)
{
    if(mesh->flattened)
    {
        if(e->buttons() == Qt::NoButton)
        {
            if(ParchmentFlattener::inLineMode())
            {
                setHoverPoint(e->x(),e->y());
                updateGL();
            }
            else if(ParchmentFlattener::markupMode == ParchmentFlattener::EditLine)
            {
                mesh->selectedLineConstraintPoint = selectLineConstraintKeypoint(e->x(),e->y());
                updateGL();
            }
            else if(ParchmentFlattener::inScaleMode())
            {
                setHoverPoint(e->x(),e->y());
                if(mesh->floatingScaleConstraint.p0fixed)
                {
                    if(mesh->floatingScaleConstraint.p1fixed)
                    {
                        mesh->floatingScaleConstraint.setP2Floating(mesh->hover_k,mesh->hover_u,mesh->hover_v);
                    }
                    else
                    {
                        mesh->floatingScaleConstraint.setP1Floating(mesh->hover_k,mesh->hover_u,mesh->hover_v);
                    }

                }
                else
                {
                    mesh->floatingScaleConstraint.setP0Floating(mesh->hover_k,mesh->hover_u,mesh->hover_v);
                }
                updateGL();
            }
        }
        else if(e->buttons() & Qt::LeftButton)
        {
            if(ParchmentFlattener::markupMode == ParchmentFlattener::EditLine)
            {
                setHoverPoint(e->x(),e->y());
                mesh->moveSelectedLineConstraintPoint(mesh->hover_k, mesh->hover_u, mesh->hover_v);
                updateGL();
                //wGlWidget->updateGL();
            }
        }
        else if(e->buttons() & Qt::RightButton)
        {
            int dy = e->y() - m_lastPoint.y();
            double zaxis[3] = {0,0,1};
            if(m_lastPoint.x() > window_width / 2)
                dy = -dy;
            rot_angle_new = rot_angle_old + (double)dy * 3.14 / 180.0;
            currentUpVec = Vector3d(qSin(rot_angle_new), qCos(rot_angle_new), 0);
            currentRightVec = currentUpVec.cross(Vector3d(0,0,1));
            igl::axis_angle_to_quat(&zaxis[0],rot_angle_new,quat_rot_new);
            igl::quat_to_mat(quat_rot_new, rotation_matrix);
            updateGL();
        }
        else if(e->buttons() & Qt::MidButton)
        {
            int delta_X = e->x() - m_lastPoint.x();
            int delta_Y = e->y() - m_lastPoint.y();
            //target_new = target + 0.001 * camz * Vector3d(delta_X,-delta_Y,0);
            target_new = target + 0.001 * camz * (delta_X*currentRightVec - delta_Y*currentUpVec);
            updateGL();
        }
    }
}

bool pointsFuzzyEqual(QPoint p1, QPoint p2, double T)
{
    double dist = QVector2D(p1-p2).length();
    return (dist <= T);
}

void FlatGLWidget::mouseReleaseEvent(QMouseEvent * e)
{
    if ((m_mouseClick) && pointsFuzzyEqual(e->pos(), m_lastPoint, 4))
    {
        mouseClickEvent(e);
    }

    if(e->button() == Qt::RightButton)
    {
        rot_angle_old = rot_angle_new;
    }
    else if(e->button() == Qt::MidButton)
    {
        target = target_new;
    }
    else if(e->button() == Qt::LeftButton)
    {
        if(ParchmentFlattener::markupMode == ParchmentFlattener::EditLine)
        {
            mesh->finishEditingLineConstraint();
        }
    }
}

void FlatGLWidget::mouseClickEvent(QMouseEvent *e)
{
    if(mesh->flattened)
    {
        if(e->button() == Qt::LeftButton)
        {
            handleClick(e->x(),e->y());
        }
        else if(e->button() == Qt::RightButton)
        {
            handleClickRight(e->x(),e->y());
        }
        updateGL();
        //wGlWidget->updateGL();
    }
}

void FlatGLWidget::wheelEvent(QWheelEvent * e)
{
    int delta = e->delta();
	if(delta > 0)
        zoomIn();
	if(delta < 0)
		zoomOut();
}

Vector3d FlatGLWidget::getSurfacePoint(int k, double s, double t)
{
    return mesh->getFlatSurfacePoint(k,s,t);
}

void getBarycentricCoordinatesOfVector(Vector3d v, Vector3d p1, Vector3d p2, Vector3d p3, double *l1, double *l2, double *l3)
{
    MatrixXd A = MatrixXd::Zero(3,3);
    A.col(0) = p1;
    A.col(1) = p2;
    A.col(2) = p3;
    Vector3d coeffs = A.inverse() * v;

    *l1 = coeffs(0);
    *l2 = coeffs(1);
    *l3 = coeffs(2);
}

template <typename T>
void jet(const T x, T & r, T & g, T & b)
{
  const double rone = 0.8;
  const double gone = 1.0;
  const double bone = 1.0;

  if(x<1/8.)
  {
    r = 0;
    g = 0;
    b = bone*(0.5+(x)/(1/8.)*0.5);
  }else if(x<3/8.)
  {
    r = 0;
    g = gone*(x-1/8.)/(3/8.-1/8.);
    b = bone;
  }else if(x<5/8.)
  {
    r = rone*(x-3/8.)/(5/8.-3/8.);
    g = gone;
    b = (bone-(x-3/8.)/(5/8.-3/8.));
  }else if(x<7/8.)
  {
    r = rone;
    g = (gone-(x-5/8.)/(7/8.-5/8.));
    b = 0;
  }else
  {
    r = (bone-(x-7/8.)/(1.-7/8.)*0.5);
    g = 0;
    b = 0;
  }
}

void mapNormalToRGB(double x,double y,double z,double &r, double &g, double &b)
{
    r = (x+1.0)/2.0;
    g = (y+1.0)/2.0;
    b = (z+1.0)/2.0;
}

void FlatGLWidget::renderMesh()
{
    if(mesh->flattened)
    {
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glEnable(GL_TEXTURE_2D);
        glEnable(GL_DEPTH_TEST);

        double minScale = 1;
        double maxScale = 1;

        double minStretch = 99999;
        double maxStretch = -99999;
        if(renderScaleField)
        {
            MatrixXd scalesY = mesh->perFaceScaleW.row(1);
            minScale = scalesY.minCoeff();
            maxScale = scalesY.maxCoeff();
        }
        else if(renderStretch)
        {
            for(int k = 0; k < mesh->F.cols(); ++k)
            {
                double areaW = mesh->getFaceArea(k);
                double areaF = mesh->getFlatFaceArea(k);
                double areaRatio = areaF / areaW;

                if(areaRatio < minStretch)
                    minStretch = areaRatio;
                if(areaRatio > maxStretch)
                    maxStretch = areaRatio;
            }
        }

        if(renderActualMesh)
        {
            glColor3d(1,1,1);
            int currentTex = -1;

            for(int k = 0; k < mesh->F.cols(); ++k)
            {
                int v0 = mesh->F(0,k);
                int v1 = mesh->F(1,k);
                int v2 = mesh->F(2,k);

                int uv0 = mesh->FTC(k,0);
                int uv1 = mesh->FTC(k,1);
                int uv2 = mesh->FTC(k,2);

                if(renderNormals)
                {
                    glDisable(GL_TEXTURE_2D);
                    Vector3d vn0 = mesh->VN.col(v0);
                    Vector3d vn1 = mesh->VN.col(v1);
                    Vector3d vn2 = mesh->VN.col(v2);

                    double r0,g0,b0;
                    double r1,g1,b1;
                    double r2,g2,b2;

                    mapNormalToRGB(vn0(0), vn0(1), vn0(2), r0, g0, b0);
                    mapNormalToRGB(vn1(0), vn1(1), vn1(2), r1, g1, b1);
                    mapNormalToRGB(vn2(0), vn2(1), vn2(2), r2, g2, b2);

                    glBegin(GL_TRIANGLES);
                    glColor3d(r0,g0,b0);
                    glVertex3d(mesh->flatV(0,v0),mesh->flatV(1,v0),mesh->flatV(2,v0));

                    glColor3d(r1,g1,b1);
                    glVertex3d(mesh->flatV(0,v1),mesh->flatV(1,v1),mesh->flatV(2,v1));

                    glColor3d(r2,g2,b2);
                    glVertex3d(mesh->flatV(0,v2),mesh->flatV(1,v2),mesh->flatV(2,v2));
                    glEnd();
                }
                else if(renderScaleField)
                {
                    glDisable(GL_TEXTURE_2D);
                    double scaleY_k = mesh->perFaceScaleW(1,k);
                    scaleY_k = (scaleY_k - minScale) / (maxScale - minScale);
                    double r,g,b;
                    jet(scaleY_k,r,g,b);
                    glColor3d(r,g,b);
                    glBegin(GL_TRIANGLES);
                    glVertex3d(mesh->flatV(0,v0),mesh->flatV(1,v0),mesh->flatV(2,v0));
                    glVertex3d(mesh->flatV(0,v1),mesh->flatV(1,v1),mesh->flatV(2,v1));
                    glVertex3d(mesh->flatV(0,v2),mesh->flatV(1,v2),mesh->flatV(2,v2));
                    glEnd();
                }
                else if(renderStretch)
                {
                    glDisable(GL_TEXTURE_2D);

                    double areaW = mesh->getFaceArea(k);
                    double areaF = mesh->getFlatFaceArea(k);

                    double areaRatio = areaF / areaW;

                    areaRatio = (areaRatio - minStretch) / (maxStretch - minStretch);

                    double r,g,b;
                    jet(areaRatio,r,g,b);
                    glColor3d(r,g,b);
                    glBegin(GL_TRIANGLES);
                    glVertex3d(mesh->flatV(0,v0),mesh->flatV(1,v0),mesh->flatV(2,v0));
                    glVertex3d(mesh->flatV(0,v1),mesh->flatV(1,v1),mesh->flatV(2,v1));
                    glVertex3d(mesh->flatV(0,v2),mesh->flatV(1,v2),mesh->flatV(2,v2));
                    glEnd();
                }
                else
                {

                    glEnable(GL_TEXTURE_2D);
                    int m = mesh->Materials(k);
                    if(m != currentTex)
                    {
                        currentTex = m;
                        if(renderEnhanced)
                            glBindTexture(GL_TEXTURE_2D, meshTextures[currentTex + numTextures]);
                        else
                            glBindTexture(GL_TEXTURE_2D, meshTextures[currentTex]);
                    }
                    glColor3d(1,1,1);
                    glBegin(GL_TRIANGLES);
                    glTexCoord2d(mesh->TC(uv0,0), mesh->TC(uv0,1));
                    glVertex3d(mesh->flatV(0,v0),mesh->flatV(1,v0),mesh->flatV(2,v0));

                    glTexCoord2d(mesh->TC(uv1,0), mesh->TC(uv1,1));
                    glVertex3d(mesh->flatV(0,v1),mesh->flatV(1,v1),mesh->flatV(2,v1));

                    glTexCoord2d(mesh->TC(uv2,0), mesh->TC(uv2,1));
                    glVertex3d(mesh->flatV(0,v2),mesh->flatV(1,v2),mesh->flatV(2,v2));
                    glEnd();
                }
            }

            glDisable(GL_TEXTURE_2D);
            glDisable(GL_DEPTH_TEST);
            glColor3d(1,0,0);

            glDisable(GL_TEXTURE_2D);
            glDisable(GL_DEPTH_TEST);
            glLineWidth(1);
        }

        if(mesh->renderCrossField)
        {
            int delta_k = 10;

            glLineWidth(2);
            for(int i = 0; i < mesh->F.cols(); i += delta_k)
            {
                int v0 = mesh->F(0,i);
                int v1 = mesh->F(1,i);
                int v2 = mesh->F(2,i);

                Vector3d p0 = mesh->flatV.col(v0);
                Vector3d p1 = mesh->flatV.col(v1);
                Vector3d p2 = mesh->flatV.col(v2);
                Vector3d vk = (p0 + p1 + p2) / 3.0;

                Vector3d rk = mesh->perFaceRightVectorsF.col(i);
                rk.normalize();

                Vector3d uk = mesh->perFaceUpVectorsF.col(i);
                uk.normalize();

                if(rk.norm() > 0)
                {
                    Vector2d scalek = mesh->perFaceScaleF.col(i);
                    double lenx = scalek(0);
                    double leny = scalek(1);

                    Vector3d vrk0 = vk;
                    Vector3d vrk1 = vk + lenx * rk;
                    Vector3d vuk0 = vk;
                    Vector3d vuk1 = vk + leny * uk;

                    glBegin(GL_LINES);                    
                    glColor3d(0,0,1);
                    glVertex3d(vrk0(0), vrk0(1), vrk0(2));
                    glVertex3d(vrk1(0), vrk1(1), vrk1(2));

                    glColor3d(0,1,0);
                    glVertex3d(vuk0(0), vuk0(1), vuk0(2));
                    glVertex3d(vuk1(0), vuk1(1), vuk1(2));

                    glEnd();
                }
            }
            glLineWidth(1);
        }
    }
}

void FlatGLWidget::renderInterpolatedMesh()
{
    glDisable(GL_BLEND);
    glDepthFunc(GL_LESS);
    if(SPIN_AROUND)
    {

        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glEnable(GL_TEXTURE_2D);
        glEnable(GL_DEPTH_TEST);
        glColor3d(1,1,1);
        int currentTex = -1;

        for(int k = 0; k < mesh->F.cols(); ++k)
        {
            int v0 = mesh->F(0,k);
            int v1 = mesh->F(1,k);
            int v2 = mesh->F(2,k);

            int uv0 = mesh->FTC(k,0);
            int uv1 = mesh->FTC(k,1);
            int uv2 = mesh->FTC(k,2);

            int m = mesh->Materials(k);
            if(m != currentTex)
            {
                currentTex = m;
                glBindTexture(GL_TEXTURE_2D, meshTextures[currentTex]);
            }

            glBegin(GL_TRIANGLES);
            glTexCoord2d(mesh->TC(uv0,0), mesh->TC(uv0,1));
            glVertex3d(mesh->V(0,v0),mesh->V(1,v0),mesh->V(2,v0));

            glTexCoord2d(mesh->TC(uv1,0), mesh->TC(uv1,1));
            glVertex3d(mesh->V(0,v1),mesh->V(1,v1),mesh->V(2,v1));

            glTexCoord2d(mesh->TC(uv2,0), mesh->TC(uv2,1));
            glVertex3d(mesh->V(0,v2),mesh->V(1,v2),mesh->V(2,v2));
            glEnd();
        }
    }
    else
    {
        MatrixXd Vinterp = mesh->interpolate3Dto2D(interpolation_t);
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glEnable(GL_TEXTURE_2D);
        glEnable(GL_DEPTH_TEST);
        glColor3d(1,1,1);
        int currentTex = -1;

        for(int k = 0; k < mesh->F.cols(); ++k)
        {
            int v0 = mesh->F(0,k);
            int v1 = mesh->F(1,k);
            int v2 = mesh->F(2,k);

            int uv0 = mesh->FTC(k,0);
            int uv1 = mesh->FTC(k,1);
            int uv2 = mesh->FTC(k,2);

            int m = mesh->Materials(k);
            if(m != currentTex)
            {
                currentTex = m;
                if(renderEnhanced)
                    glBindTexture(GL_TEXTURE_2D, meshTextures[currentTex + numTextures]);
                else
                    glBindTexture(GL_TEXTURE_2D, meshTextures[currentTex]);
            }

            glBegin(GL_TRIANGLES);
            glTexCoord2d(mesh->TC(uv0,0), mesh->TC(uv0,1));
            glVertex3d(Vinterp(0,v0),Vinterp(1,v0),Vinterp(2,v0));

            glTexCoord2d(mesh->TC(uv1,0), mesh->TC(uv1,1));
            glVertex3d(Vinterp(0,v1),Vinterp(1,v1),Vinterp(2,v1));

            glTexCoord2d(mesh->TC(uv2,0), mesh->TC(uv2,1));
            glVertex3d(Vinterp(0,v2),Vinterp(1,v2),Vinterp(2,v2));
            glEnd();
        }

        if(renderConstraints)
        {
            glDisable(GL_TEXTURE_2D);
            glEnable(GL_DEPTH_TEST);
//            glPointSize(5);
//            for(int j = 0; j < mesh->lineConstraints.size(); ++j)
//            {
//                if(mesh->lineConstraints[j].isActive())
//                {
//                    glColor3d(0,0,1);
//                    LineConstraint lc = mesh->lineConstraints[j];
//                    for(int i = 0; i < lc.size(); ++i)
//                    {
//                        int ki = lc.new_faces[i];
//                        double si = lc.new_sts(0,i);
//                        double ti = lc.new_sts(1,i);

//                        int vi0 = mesh->F(0,ki);
//                        int vi1 = mesh->F(1,ki);
//                        int vi2 = mesh->F(2,ki);

//                        Vector3d v0 = Vinterp.col(vi0);
//                        Vector3d v1 = Vinterp.col(vi1);
//                        Vector3d v2 = Vinterp.col(vi2);

//                        Vector3d p = (1 - si - ti) * v0 + si * v1 + ti * v2;
//                        p(2) = p(2) + 0.01;

//                        glPointSize(5);
//                        glBegin(GL_POINTS);
//                        glVertex3d(p(0),p(1),p(2));
//                        glEnd();
//                    }
//                }
//            }

            glLineWidth(10);
            for(int j = 0; j < mesh->lineConstraints.size(); ++j)
            {
                if(mesh->lineConstraints[j].isActive())
                {
                    glColor3d(0,0,1);
                    LineConstraint lc = mesh->lineConstraints[j];
                    glBegin(GL_LINE_STRIP);
                    for(int i = 0; i < lc.size(); ++i)
                    {
                        int ki = lc.new_faces[i];
                        double si = lc.new_sts(0,i);
                        double ti = lc.new_sts(1,i);

                        int vi0 = mesh->F(0,ki);
                        int vi1 = mesh->F(1,ki);
                        int vi2 = mesh->F(2,ki);

                        Vector3d v0 = Vinterp.col(vi0);
                        Vector3d v1 = Vinterp.col(vi1);
                        Vector3d v2 = Vinterp.col(vi2);

                        Vector3d p = (1 - si - ti) * v0 + si * v1 + ti * v2;
                        p(2) = p(2) + 0.01;

                        glVertex3d(p(0),p(1),p(2));
                    }
                    glEnd();
                }
            }

            glLineWidth(8);
            for(int i = 0; i < mesh->scaleConstraints2.size(); ++i)
            {
                ScaleConstraint sc = mesh->scaleConstraints2[i];
                if(!sc.isActive())
                    continue;

                if(sc.face0 > 0 && sc.face1 > 0)
                {
                    int k0 = sc.face0;      int k1 = sc.face1;
                    double u0 = sc.u0;      double u1 = sc.u1;
                    double v0 = sc.v0;      double v1 = sc.v1;

                    int v00 = mesh->F(0,k0);    int v10 = mesh->F(0,k1);
                    int v01 = mesh->F(1,k0);    int v11 = mesh->F(1,k1);
                    int v02 = mesh->F(2,k0);    int v12 = mesh->F(2,k1);

                    Vector3d p0 = Vinterp.col(v00);     Vector3d q0 = Vinterp.col(v10);
                    Vector3d p1 = Vinterp.col(v01);     Vector3d q1 = Vinterp.col(v11);
                    Vector3d p2 = Vinterp.col(v02);     Vector3d q2 = Vinterp.col(v12);

                    Vector3d pbot = (1 - u0 - v0) * p0 + u0 * p1 + v0 * p2;
                    Vector3d ptop = (1 - u1 - v1) * q0 + u1 * q1 + v1 * q2;

                    Vector3d pmid = (ptop + pbot) / 2.0;
                    pmid(2) = pmid(2) + 0.01;

                    if(i == mesh->selectedScaleConstraint)
                        glColor3d(1,0,0);
                    else
                        glColor3d(0,0,1);

                    double rad = (ptop - pbot).norm() / 2;
                    DrawCircle(pmid(0), pmid(1), pmid(2), rad, 20);
                }
            }
        }
    }
    glLineWidth(1);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_DEPTH_TEST);
}

void FlatGLWidget::renderOriginalMesh()
{
    if(texturesLoaded)
    {
        glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
        glEnable(GL_TEXTURE_2D);
        glEnable(GL_DEPTH_TEST);
        glColor3d(1,1,1);
        int currentTex = -1;

        double minScale = 1;
        double maxScale = 1;

        double minScale2 = 1;
        double maxScale2 = 1;
        if(renderScaleField)
        {
            MatrixXd scalesY = mesh->perFaceScaleW.row(1);
            qDebug() << scalesY.rows() << ", " << scalesY.cols();
            minScale = scalesY.minCoeff();
            maxScale = scalesY.maxCoeff();
            qDebug() << minScale << ", " << maxScale;

            MatrixXd scalesY2 = mesh->perVertexScaleW.row(1);
            qDebug() << scalesY2.rows() << ", " << scalesY2.cols();
            minScale2 = scalesY2.minCoeff();
            maxScale2 = scalesY2.maxCoeff();
            qDebug() << minScale2 << ", " << maxScale2;
        }

        for(int k = 0; k < mesh->F.cols(); ++k)
        {
            int v0 = mesh->F(0,k);
            int v1 = mesh->F(1,k);
            int v2 = mesh->F(2,k);

            int uv0 = mesh->FTC(k,0);
            int uv1 = mesh->FTC(k,1);
            int uv2 = mesh->FTC(k,2);

            int m = mesh->Materials(k);
            if(m != currentTex)
            {
                currentTex = m;
                glBindTexture(GL_TEXTURE_2D, meshTextures[currentTex]);
            }
            else if(renderScaleField)
            {
                glDisable(GL_TEXTURE_2D);
                double scaleY_k = mesh->perFaceScaleW(1,k);

                scaleY_k = (scaleY_k - minScale) / (maxScale - minScale);
                scaleY_k = scaleY_k * 0.6 + 0.3;

                double r,g,b;
                jet(scaleY_k,r,g,b);
                glColor3d(r,g,b);

                glBegin(GL_TRIANGLES);
                glVertex3d(mesh->V(0,v0),mesh->V(1,v0),mesh->V(2,v0));
                glVertex3d(mesh->V(0,v1),mesh->V(1,v1),mesh->V(2,v1));
                glVertex3d(mesh->V(0,v2),mesh->V(1,v2),mesh->V(2,v2));
                glEnd();
            }
            else
            {
                glBegin(GL_TRIANGLES);
                glTexCoord2d(mesh->TC(uv0,0), mesh->TC(uv0,1));
                glVertex3d(mesh->V(0,v0),mesh->V(1,v0),mesh->V(2,v0));

                glTexCoord2d(mesh->TC(uv1,0), mesh->TC(uv1,1));
                glVertex3d(mesh->V(0,v1),mesh->V(1,v1),mesh->V(2,v1));

                glTexCoord2d(mesh->TC(uv2,0), mesh->TC(uv2,1));
                glVertex3d(mesh->V(0,v2),mesh->V(1,v2),mesh->V(2,v2));
                glEnd();
            }
        }

        glDisable(GL_TEXTURE_2D);
        glDisable(GL_DEPTH_TEST);

        if(mesh->renderCrossField)
        {
            int delta_k = 12;

            /*
            if(mesh->loading_flow)
            {
                for(int i = 0; i < mesh->Vuntouched.cols(); i += delta_k)
                {
                    Vector3d vk = mesh->Vuntouched.col(i);
                    Vector3d rk = mesh->perVertexRightVectorsW.col(i);
                    rk.normalize();

                    if(rk.norm() > 0)
                    {
                        double lenx = 0.1;

                        Vector3d vrk0 = vk;
                        Vector3d vrk1 = vk + lenx/2 * rk;

                        glBegin(GL_LINES);

                        glColor3d(1,0,0);
                        glVertex3d(vrk0(0), vrk0(1), vrk0(2));
                        glVertex3d(vrk1(0), vrk1(1), vrk1(2));


                        glEnd();

                    }
                }
                return;
            }
            */

            glLineWidth(4);
            for(int i = 0; i < mesh->F.cols(); i += delta_k)
            {
                int v0 = mesh->F(0,i);
                int v1 = mesh->F(1,i);
                int v2 = mesh->F(2,i);

                Vector3d p0 = mesh->V.col(v0);
                Vector3d p1 = mesh->V.col(v1);
                Vector3d p2 = mesh->V.col(v2);
                Vector3d vk = (p0 + p1 + p2) / 3.0;

                Vector3d rk = mesh->perFaceRightVectorsW.col(i);
                rk.normalize();

                Vector3d uk = mesh->perFaceUpVectorsW.col(i);
                uk.normalize();

                if(rk.norm() > 0)
                {
                    Vector2d scalek = mesh->perFaceScaleW.col(i);
                    double lenx = scalek(0);
                    double leny = scalek(1);

                    //if(lenx >= 1)
                    //    lenx *= 0.025;
                    //if(leny >= 1)
                    //    leny *= 0.025;

                    lenx *= 2;
                    leny *= 2;

                    //Vector3d vrk0 = vk - len/2 * rk;
                    Vector3d vrk0 = vk;
                    Vector3d vrk1 = vk + lenx/2 * rk;
                    //Vector3d vuk0 = vk - len/2 * uk;
                    Vector3d vuk0 = vk;
                    Vector3d vuk1 = vk + leny/2 * uk;

                    glBegin(GL_LINES);

                    glColor3d(0,0,1);
                    glVertex3d(vrk0(0), vrk0(1), vrk0(2));
                    glVertex3d(vrk1(0), vrk1(1), vrk1(2));

                    glColor3d(0,1,0);
                    glVertex3d(vuk0(0), vuk0(1), vuk0(2));
                    glVertex3d(vuk1(0), vuk1(1), vuk1(2));

                    glEnd();

//                    glPointSize(3);
//                    glBegin(GL_POINTS);
//                    if(mesh->isScaleConstrained(v0) < 0 || mesh->isScaleConstrained(v1) < 0 || mesh->isScaleConstrained(v2) < 0)
//                        glColor3d(0,0,1);
//                    else
//                        glColor3d(1,0,0);
//                    glVertex3d(vrk0(0), vrk0(1), vrk0(2));
//                    glEnd();
                }
            }
            glLineWidth(1);
        }
    }
}

void FlatGLWidget::renderLineConstraints()
{
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_DEPTH_TEST);
    glPointSize(5);

    for(int j = 0; j < mesh->lineConstraints.size(); ++j)
    {
        if(mesh->lineConstraints[j].isActive())
        {
            if(ParchmentFlattener::markupMode == ParchmentFlattener::EditLine && j == mesh->selectedLineConstraint)
            {
                glColor3d(1,0,0);
            }
            else
            {
                glColor3d(0,0,1);
            }
            LineConstraint lc = mesh->lineConstraints[j];
            for(int i = 0; i < lc.size(); ++i)
            {
                int ki = lc.new_faces[i];
                double si = lc.new_sts(0,i);
                double ti = lc.new_sts(1,i);
                Vector3d p = getSurfacePoint(ki,si,ti);
                if(i == mesh->selectedLineConstraintPoint && j == mesh->selectedLineConstraint)
                    glPointSize(10);
                else
                    glPointSize(5);
                glBegin(GL_POINTS);
                glVertex3d(p(0),p(1),p(2));
                glEnd();
            }
        }
    }

    glLineWidth(5);
    for(int j = 0; j < mesh->lineConstraints.size(); ++j)
    {
        if(mesh->lineConstraints[j].isActive())
        {
            if(ParchmentFlattener::markupMode == ParchmentFlattener::EditLine && j == mesh->selectedLineConstraint)
            {
                glColor3d(1,0,0);
            }
            else
            {
                glColor3d(0,0,1);
            }
            LineConstraint lc = mesh->lineConstraints[j];
            glBegin(GL_LINE_STRIP);
            for(int i = 0; i < lc.size(); ++i)
            {
                int ki = lc.new_faces[i];
                double si = lc.new_sts(0,i);
                double ti = lc.new_sts(1,i);
                Vector3d p = getSurfacePoint(ki,si,ti);
                glVertex3d(p(0),p(1),p(2));
            }
            glEnd();
        }
    }
    glLineWidth(1);

    if(ParchmentFlattener::inLineMode() && mesh->floatingLineConstraint.size() > 0)
    {
        glBegin(GL_POINTS);
        glColor3d(1,0,0);
        for(int i = 0; i < mesh->floatingLineConstraint.size(); ++i)
        {
            int ki = mesh->floatingLineConstraint.faces[i];
            double si = mesh->floatingLineConstraint.sts(0,i);
            double ti = mesh->floatingLineConstraint.sts(1,i);
            Vector3d p = getSurfacePoint(ki,si,ti);
            glVertex3d(p(0),p(1),p(2));
        }
//        int ki = mesh->hover_k;
//        double si = mesh->hover_u;
//        double ti = mesh->hover_v;
//        Vector3d p = getSurfacePoint(ki,si,ti);
//        glVertex3d(p(0),p(1),p(2));
        glEnd();

        glLineWidth(4);
        glBegin(GL_LINE_STRIP);
        glColor3d(1,0,0);
        for(int i = 0; i < mesh->floatingLineConstraint.size(); ++i)
        {
            int ki = mesh->floatingLineConstraint.faces[i];
            double si = mesh->floatingLineConstraint.sts(0,i);
            double ti = mesh->floatingLineConstraint.sts(1,i);
            Vector3d p = getSurfacePoint(ki,si,ti);
            glVertex3d(p(0),p(1),p(2));
        }
        glLineWidth(1);
        if(mesh->floatingLineConstraint.size() > 0)
        {
            int ki = mesh->hover_k;
            double si = mesh->hover_u;
            double ti = mesh->hover_v;
            Vector3d p = getSurfacePoint(ki,si,ti);
            glVertex3d(p(0),p(1),p(2));
        }
        glEnd();

        // RENDER SUGGESTION TAIL
        glBegin(GL_POINTS);
        glColor3d(0,1,0);
        for(int i = 0; i < mesh->floatingLineConstraint.suggested_faces.size(); ++i)
        {
            int ki = mesh->floatingLineConstraint.suggested_faces[i];
            double si = mesh->floatingLineConstraint.suggested_sts(0,i);
            double ti = mesh->floatingLineConstraint.suggested_sts(1,i);
            Vector3d p = getSurfacePoint(ki,si,ti);
            glVertex3d(p(0),p(1),p(2));
        }
        glEnd();

        glLineWidth(4);
        glBegin(GL_LINE_STRIP);
        glColor3d(0,1,0);
        int endi = mesh->floatingLineConstraint.size() - 1;
        int kend = mesh->floatingLineConstraint.faces[endi];
        double send = mesh->floatingLineConstraint.sts(0,endi);
        double tend = mesh->floatingLineConstraint.sts(1,endi);
        Vector3d pend = getSurfacePoint(kend,send,tend);
        glVertex3d(pend(0),pend(1),pend(2));
        for(int i = 0; i < mesh->floatingLineConstraint.suggested_faces.size(); ++i)
        {
            int ki = mesh->floatingLineConstraint.suggested_faces[i];
            double si = mesh->floatingLineConstraint.suggested_sts(0,i);
            double ti = mesh->floatingLineConstraint.suggested_sts(1,i);
            Vector3d p = getSurfacePoint(ki,si,ti);
            glVertex3d(p(0),p(1),p(2));
        }
        glEnd();
    }
    glLineWidth(1);
    glPointSize(1);
}

void FlatGLWidget::renderLineConstraintsW()
{
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST);
    glPointSize(5);

    for(int j = 0; j < mesh->lineConstraints.size(); ++j)
    {
        if(mesh->lineConstraints[j].isActive())
        {
            glColor3d(0,0,1);

            LineConstraint lc = mesh->lineConstraints[j];
            for(int i = 0; i < lc.size(); ++i)
            {
                int ki = lc.new_faces[i];
                double si = lc.new_sts(0,i);
                double ti = lc.new_sts(1,i);
                Vector3d p = mesh->getWrinkledSurfacePoint(ki,si,ti);
                p(2) = p(2) + 0.01;
                glPointSize(5);
                glBegin(GL_POINTS);
                glVertex3d(p(0),p(1),p(2));
                glEnd();
            }
        }
    }

    glLineWidth(5);
    for(int j = 0; j < mesh->lineConstraints.size(); ++j)
    {
        if(mesh->lineConstraints[j].isActive())
        {
            glColor3d(0,0,1);
            LineConstraint lc = mesh->lineConstraints[j];
            glBegin(GL_LINE_STRIP);
            for(int i = 0; i < lc.size(); ++i)
            {
                int ki = lc.new_faces[i];
                double si = lc.new_sts(0,i);
                double ti = lc.new_sts(1,i);
                Vector3d p = mesh->getWrinkledSurfacePoint(ki,si,ti);
                p(2) = p(2) + 0.01;
                glVertex3d(p(0),p(1),p(2));
            }
            glEnd();
        }
    }
    glLineWidth(1);
    glPointSize(1);
}

void FlatGLWidget::renderScaleConstraints()
{
    for(int i = 0; i < mesh->scaleConstraints2.size(); ++i)
    {
        if(i == mesh->selectedScaleConstraint)
            glColor3d(1,0,0);
        else
            glColor3d(0,0,1);
        renderScaleConstraint(mesh->scaleConstraints2[i]);
    }
    glColor3d(1,0,0);
    renderScaleConstraint(mesh->floatingScaleConstraint);
}

void FlatGLWidget::renderScaleConstraintsW()
{
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_DEPTH_TEST);
    glColor3d(0,0,1);
    glLineWidth(5);
    for(int i = 0; i < mesh->scaleConstraints2.size(); ++i)
    {
        ScaleConstraint sc = mesh->scaleConstraints2[i];
        if(!sc.isActive())
            continue;
        if(sc.face0 > 0 && sc.face1 > 0)
        {
            Vector3d p0 = mesh->getWrinkledSurfacePoint(sc.face0,sc.u0,sc.v0);
            Vector3d p1 = mesh->getWrinkledSurfacePoint(sc.face1,sc.u1,sc.v1);
            Vector3d pmid = (p0 + p1) / 2.0;
            pmid(2) = pmid(2) + 0.01;

            if(i == mesh->selectedScaleConstraint)
                glColor3d(1,0,0);
            else
                glColor3d(0,0,1);

            double rad = (p0 - p1).norm() / 2;
            DrawCircle(pmid(0), pmid(1), pmid(2), rad, 30);
        }
    }
    glLineWidth(1);
}

QImage FlatGLWidget::renderScreenToQImage(int w, int h)
{
    QGLPixelBuffer *pixBuffer = new QGLPixelBuffer(w, h, QGLFormat::defaultFormat(), this);
    pixBuffer->makeCurrent();
    int old_window_width = window_width;
    int old_window_height = window_height;
    window_width = w;
    window_height = h;
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    setProjectionMatrixGL();
    setModelViewMatrixGL();

    //renderOriginalMesh();

    if(renderInterpolation)
    {
        renderInterpolatedMesh();
    }
    else
    {
        renderMesh();
        if(renderConstraints)
        {
            //renderLineConstraintsW();
            //renderScaleConstraintsW();
            //renderLineConstraints();
            //renderScaleConstraints();
        }
    }

    bool ok;
    QString imName = QInputDialog::getText(0,tr("Image Name"), tr("Enter image name and file extension (eg. image.png)"),QLineEdit::Normal,QString(),&ok);

    QImage im = pixBuffer->toImage();

    if(!ok)
        return im;

    delete pixBuffer;

    QString filename = ParchmentFlattener::meshDirectory;
    filename.append(imName);

    im.save(filename);

    makeCurrent();
    window_width = old_window_width;
    window_height = old_window_height;
    setProjectionMatrixGL();
    setModelViewMatrixGL();

    return im;
}

QImage FlatGLWidget::renderTargetToQImage(Vector3d render_target, int w, int h)
{
    QGLPixelBuffer *pixBuffer = new QGLPixelBuffer(w, h, QGLFormat::defaultFormat(), this);
    pixBuffer->makeCurrent();
    int old_window_width = window_width;
    int old_window_height = window_height;
    window_width = w;
    window_height = h;
    setProjectionMatrixGL();
    setModelViewMatrixGL(render_target);
    renderMesh();
    QImage im = pixBuffer->toImage();
    delete pixBuffer;

    makeCurrent();
    window_width = old_window_width;
    window_height = old_window_height;
    setProjectionMatrixGL();
    setModelViewMatrixGL();

    return im;
}

void FlatGLWidget::renderScaleConstraint(ScaleConstraint sc)
{
    if(!sc.isActive())
        return;

    if(sc.face0 > 0)
    {
        glLineWidth(5);
        if(sc.face1 > 0)
        {
            Vector3d p0 = getSurfacePoint(sc.face0,sc.u0,sc.v0);
            Vector3d p1 = getSurfacePoint(sc.face1,sc.u1,sc.v1);
            Vector3d pmid = (p0 + p1) / 2.0;
            //if(sc.face2 > 0)
            if(sc.scale(0) != sc.scale(1) || sc.face2 >= 0)
            {
                Vector3d p2;
                if(sc.face2 >= 0)
                    p2 = getSurfacePoint(sc.face2,sc.u2,sc.v2);
                double rady = (p1 - pmid).norm();
                Vector3d yaxis = (p1 - pmid);
                yaxis.normalize();
                Vector3d xaxis = yaxis.cross(Vector3d(0,0,1));
                xaxis.normalize();
                double radx = rady;
                if(sc.face2 > 0)
                    radx = (p2 - pmid).dot(xaxis);
                else
                    radx = rady * sc.scale(0)/sc.scale(1);
                radx = qAbs(radx);

                DrawElipse(pmid, xaxis, yaxis, radx, rady, 30);

            }
            else
            {
                double rad = (p0 - p1).norm() / 2;
                DrawCircle(pmid(0), pmid(1), pmid(2), rad, 30);
            }
        }
        else
        {
            glPointSize(5);
            glBegin(GL_POINTS);
            Vector3d p0 = getSurfacePoint(sc.face0,sc.u0,sc.v0);
            glVertex3d(p0(0),p0(1),p0(2));
            glEnd();
            glPointSize(1);
        }
        glLineWidth(1);
    }

    return;

    if(sc.p0fixed && sc.p1fixed)
    {
        Vector3d p0 = getSurfacePoint(sc.face0,sc.u0,sc.v0);
        Vector3d p1 = getSurfacePoint(sc.face1,sc.u1,sc.v1);
        //Vector3d pmid = getSurfacePoint(sc.face,sc.u,sc.v);
        Vector3d pmid = (p0 + p1) / 2.0;

        double rad = (p0 - p1).norm() / 2;

        /*
        Vector3d uvec = p1 - p0;
        uvec.normalize();
        Vector3d rvec = nave.cross(uvec);
        rvec.normalize();

        Vector3d p2 = p0 - dist/2 * rvec;
        Vector3d p3 = p0 + dist/2 * rvec;
        Vector3d p4 = p1 - dist/2 * rvec;
        Vector3d p5 = p1 + dist/2 * rvec;

        glBegin(GL_LINES);
        glVertex3d(p0(0),p0(1),p0(2));
        glVertex3d(p1(0),p1(1),p1(2));
        glVertex3d(p2(0),p2(1),p2(2));
        glVertex3d(p3(0),p3(1),p3(2));
        glVertex3d(p4(0),p4(1),p4(2));
        glVertex3d(p5(0),p5(1),p5(2));
        glEnd();
        */

        DrawCircle(pmid(0), pmid(1), pmid(2), rad, 30);
    }
    else
    {
        glPointSize(5);
        glBegin(GL_POINTS);
        if(sc.p0fixed)
        {
            Vector3d p0 = getSurfacePoint(sc.face0,sc.u0,sc.v0);
            glVertex3d(p0(0),p0(1),p0(2));
        }
        if(sc.p1fixed)
        {
            Vector3d p1 = getSurfacePoint(sc.face1,sc.u1,sc.v1);
            glVertex3d(p1(0),p1(1),p1(2));
        }
        glEnd();
        glPointSize(1);
    }
}

bool FlatGLWidget::clickOnMesh(double winX, double winY, int *k, double *s, double *t, Vector3d *p)
{
    GLfloat winZ_gl = 0;
    glReadPixels( int(winX), int(winY), 1, 1, GL_DEPTH_COMPONENT, GL_FLOAT, &winZ_gl );
    double winZ = (double)winZ_gl;
    znear = Z_NEAR_PROJ;
    setProjectionMatrixGL();
    znear = Z_NEAR_MAIN;
    setModelViewMatrixGL();
    Vector3d Q = fullUnProjectionGL(winX,winY,winZ);
    Vector3d O = fullUnProjectionGL(winX,winY,0.0);
    Vector3d D = Q-O;
    double dist = 0;
    bool is_intersect = mesh->intersectMeshWithRayFlat(O,D,&dist,k,s,t);
    if(is_intersect)
        *p = O + (dist) * D;
    else
        *p = Vector3d(0,0,0);
    return is_intersect;
}

int FlatGLWidget::selectSuggestionTailPoint(int mouseX, int mouseY)
{
    int winX = mouseX;
    int winY = window_height - mouseY;
    QPointF pm((qreal)winX, (qreal)winY);

    LineConstraint lc = mesh->floatingLineConstraint;

    for(int i = 0; i < lc.suggested_faces.size(); ++i)
    {
        Vector3d sp = getSurfacePoint(lc.suggested_faces[i], lc.suggested_sts(0,i), lc.suggested_sts(1,i));
        QPointF pv = fullProjectionGL(sp(0), sp(1), sp(2));

        double dist = QLineF(pm,pv).length();

        if(dist <= CLICK_THRESH)
        {
            return i;
        }
    }
    return -1;
}

void FlatGLWidget::handleClick(int winX, int winY)
{
    if(ParchmentFlattener::inLineMode())
    {
        int selectedTailPoint = selectSuggestionTailPoint(winX,winY);

        if(selectedTailPoint >= 0)
        {
            mesh->floatingLineConstraint.acceptSuggestions(selectedTailPoint);
            generateSuggestionTailDynamic();
        }
        else
        {
            int k = -1;
            double s = 0;
            double t = 0;
            Vector3d p;

            int winX2 = winX;
            int winY2 = window_height - winY;

            bool isIntersect = clickOnMesh(winX2,winY2,&k,&s,&t,&p);

            if(isIntersect)
            {
                //qDebug() << "click line constraint";
                mesh->clickLineConstraint(k,s,t);
                generateSuggestionTailDynamic();
            }
        }
        updateGL();
    }
    else if(ParchmentFlattener::inScaleMode())
    {
        int k = -1;
        double s = 0;
        double t = 0;
        Vector3d p;

        int winX2 = winX;
        int winY2 = window_height - winY;

        bool isIntersect = clickOnMesh(winX2,winY2,&k,&s,&t,&p);

        if(isIntersect)
        {
            mesh->clickScaleConstraint(k,s,t,false);
        }
        updateGL();
    }
}

void FlatGLWidget::handleClickRight(int winX, int winY)
{
    if(ParchmentFlattener::inLineMode())
    {
        //qDebug() << "finish line constraint";
        mesh->finishLineConstraint(ParchmentFlattener::verticalLines);
    }
    else if(ParchmentFlattener::inScaleMode())
    {
        //qDebug() << "finish scale constrant Y. Begin scale constraint X axis";
        int k = -1;
        double s = 0;
        double t = 0;
        Vector3d p;

        int winX2 = winX;
        int winY2 = window_height - winY;

        bool isIntersect = clickOnMesh(winX2,winY2,&k,&s,&t,&p);

        if(isIntersect)
        {
            mesh->clickScaleConstraint(k,s,t,true);
        }
        updateGL();
    }
}

void FlatGLWidget::smoothingIterationWasDone()
{
    resetRotation();
    updateGL();
}

void FlatGLWidget::flatteningIterationWasDone()
{
    resetRotation();
    updateGL();
}

void FlatGLWidget::generateSuggestionTail(int numSteps)
{
    if(mesh->floatingLineConstraint.size() == 0)
        return;

    setModelViewMatrixGL();
    double modelview[16];
    glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
    double rx = modelview[0];
    double ry = modelview[4];
    Vector3d step(rx,ry,0);

    if(ParchmentFlattener::verticalLines)
    {
        rx = -modelview[1];
        ry = -modelview[5];
        step = Vector3d(rx,ry,0);
        return;
    }

    if(numSteps < 0)
        numSteps = DEFAULT_SUGGESTION_TAIL_LENGTH;

    if(ParchmentFlattener::verticalLines)
        numSteps = 1;

    Vector3d target_old = target_new;
    double camz_old = camz;
    camz = camz / 5;
    for(int stepNum = 0; stepNum < numSteps; ++stepNum)
    {
        Vector3d pi, pj;
        if(mesh->floatingLineConstraint.suggested_faces.size() == 0)
        {
            int i = mesh->floatingLineConstraint.size() - 1;
            int ki = mesh->floatingLineConstraint.new_faces[i];
            double ui = mesh->floatingLineConstraint.new_sts(0,i);
            double vi = mesh->floatingLineConstraint.new_sts(1,i);
            pi = mesh->getFlatSurfacePoint(ki,ui,vi);

            if(mesh->floatingLineConstraint.size() > 1)
            {
                int j = mesh->floatingLineConstraint.size() - 2;
                int kj = mesh->floatingLineConstraint.new_faces[j];
                double uj = mesh->floatingLineConstraint.new_sts(0,j);
                double vj = mesh->floatingLineConstraint.new_sts(1,j);
                pj = mesh->getFlatSurfacePoint(kj,uj,vj);
                step = pi - pj;
            }
            else
            {
                step = Vector3d(rx,ry,0);
            }
        }
        else
        {
            int i = mesh->floatingLineConstraint.suggested_faces.size() - 1;
            int ki = mesh->floatingLineConstraint.suggested_faces[i];
            double ui = mesh->floatingLineConstraint.suggested_sts(0,i);
            double vi = mesh->floatingLineConstraint.suggested_sts(1,i);
            pi = mesh->getFlatSurfacePoint(ki,ui,vi);

            if(mesh->floatingLineConstraint.suggested_faces.size() > 1)
            {
                int j = mesh->floatingLineConstraint.suggested_faces.size() - 2;
                int kj = mesh->floatingLineConstraint.suggested_faces[j];
                double uj = mesh->floatingLineConstraint.suggested_sts(0,j);
                double vj = mesh->floatingLineConstraint.suggested_sts(1,j);
                pj = mesh->getFlatSurfacePoint(kj,uj,vj);
                step = pi - pj;
            }
            else
            {
                int j = mesh->floatingLineConstraint.size() - 1;
                int kj = mesh->floatingLineConstraint.new_faces[j];
                double uj = mesh->floatingLineConstraint.new_sts(0,j);
                double vj = mesh->floatingLineConstraint.new_sts(1,j);
                pj = mesh->getFlatSurfacePoint(kj,uj,vj);
                step = pi - pj;
            }
        }

        step.normalize();
        step = step * 0.05;
        Vector3d pstep = pi + step;

        //qDebug() << "pi: " << pi(0) << "," << pi(1) << "," << pi(2);
        //qDebug() << "step: " << step(0) << "," << step(1) << "," << step(2);
        //qDebug() << "pstep: " << pstep(0) << "," << pstep(1) << "," << pstep(2);

        setTarget(pstep);

        setModelViewMatrixGL();
        //qDebug() << "target_new: " << target_new(0) << "," << target_new(1) << "," << target_new(2) << "\n";

        int imW = 101;
        int imH = 400;
        if(ParchmentFlattener::verticalLines)
        {
            imW = 400;
            double ratio = (double)window_height / (double)window_width;
            imH = imW * ratio;
        }
        QImage im = renderTargetToQImage(pstep,imW, imH);
        if(ParchmentFlattener::verticalLines)
        {
            int midH = imH / 2;
            if(imH % 2 > 0)
                midH += 1;
            int topH = midH - 50;

            im = im.copy(0,topH,imW,101);
        }
        Vector2i lineLocalization = automationEngine->localizeLineInImage(im);
        int px = lineLocalization(0);
        int py = lineLocalization(1);
        double winX, winY;
        if(ParchmentFlattener::verticalLines)
        {
            double widthScale = (double)window_width/(double)imW;
            winX = px * widthScale;
            winY = (double)window_height/2.0;
        }
        else
        {
            double heightScale = (double)window_height/(double)imH;
            winY = py * heightScale;
            winY = window_height - winY;
            winX = (double)window_width/2.0;
        }

        int k = -1;
        double u = -1;
        double v = -1;
        Vector3d newpoint;

        clickOnMesh(winX,winY,&k,&u,&v,&newpoint);

        if(k >= 0)
        {
            mesh->floatingLineConstraint.addSuggestionPoint(k,u,v);
        }
        else
        {
            break;
        }
    }
    setTarget(target_old);
    camz = camz_old;
    setModelViewMatrixGL();
    updateGL();
}

Mat derivativeOfGaussian2(int size, double sigma)
{
    Mat G = Mat(size,size,CV_64F);
    int rad = size/2;
    double sum = 0;
    for(int i = 0; i < size; ++i)
    {
        int x = i - rad;
        for(int j = 0; j < size; ++j)
        {
            int y = j - rad;
            double Gxy = y * exp( -(x*x + y*y) / (2*sigma*sigma) );
            G.at<double>(j,i) = Gxy;
            sum += Gxy;
        }
    }
    for(int i = 0; i < size; ++i)
    {
        for(int j = 0; j < size; ++ j)
        {
            G.at<double>(j,i) = G.at<double>(j,i) / sum;
        }
    }

    return G;
}

std::vector<QPoint> FlatGLWidget::smoothBaseline(std::vector<QPoint> ps1, int n, int from, int to)
{
    std::vector<QPoint> psold;
    std::vector<QPoint> psnew;

    if(from < 0)
        from = 0;

    if(to < 0)
        to = ps1.size() - 1;

    psold = ps1;
    for(int j = 0; j < n; ++j)
    {
        psnew.clear();
        for(int i = from; i <= to; ++i)
        {
            QPoint p = psold[i];
            QPoint pright, pleft;
            QPoint newp;
            if(i == 0)
            {
                pright = psold[i+1];
                pleft = (p - pright) + p;
            }
            else if(i == psold.size() - 1)
            {
                pleft = psold[i-1];
                pright = (p - pleft) + p;
            }
            else
            {
                pleft = psold[i-1];
                pright = psold[i+1];
            }
            newp = (pright + pleft)/2.0;
            psnew.push_back(newp);
        }
        psold = psnew;
    }
    return psnew;
}

void FlatGLWidget::generateSuggestionTailDynamic()
{
    if(!DO_SUGGESTION_TAILS)
        return;

    if(ParchmentFlattener::verticalLines)
        return;

    int i = mesh->floatingLineConstraint.size() - 1;
    int ki = mesh->floatingLineConstraint.new_faces[i];
    double ui = mesh->floatingLineConstraint.new_sts(0,i);
    double vi = mesh->floatingLineConstraint.new_sts(1,i);
    Vector3d pi = mesh->getFlatSurfacePoint(ki,ui,vi);

    QPointF startPixel = fullProjectionGL(pi(0), pi(1), pi(2));
    startPixel.setY(window_height - startPixel.y());

    double scale_factor = 0.5;
    int imW = window_width * scale_factor;
    int imH = window_height * scale_factor;
    startPixel = startPixel * scale_factor;

    //QImage im = renderTargetToQImage(pstep,imW, imH);
    QImage im = renderScreenToQImage(imW, imH);
    Mat mat(im.height(), im.width(), CV_8UC4, im.bits(), im.bytesPerLine());
    Mat Temp;
    mat.convertTo(Temp,CV_8U);
    Mat img;
    cv::cvtColor(Temp, img, CV_BGRA2GRAY);
    Mat dImg = img.clone();

    int kernelHeight = 100 * scale_factor;
    int kernelSD = 10 * scale_factor;
    Mat wideKernel = derivativeOfGaussian2(kernelHeight,kernelSD);
    filter2D(img,dImg,5,wideKernel);

    cv::normalize(wideKernel, wideKernel, 0, 255, NORM_MINMAX, CV_32SC1);
    cv::normalize(dImg, dImg, 0, 255, NORM_MINMAX, CV_32SC1);
    //cv::imwrite("kernel.jpg",wideKernel);
    //cv::imwrite("mainim.jpg",img);
    //cv::imwrite("gradim.jpg",dImg);

    std::vector<QPoint> points;

    Mat costMatrix = Mat::zeros(dImg.rows, dImg.cols, CV_64FC1);
    Mat backMatrix = Mat::zeros(dImg.rows, dImg.cols, CV_32SC1);

    int start_col = (int)startPixel.x();
    int start_row = (int)startPixel.y();

    for(int c = 0; c < dImg.cols; ++c)
    {
        for(int r = 0; r < dImg.rows; ++r)
        {
            double bestCost = 99999;
            int bestBack = 0;

            // Compute unary cost
            int dImgVal = dImg.at<int>(r,c);
            double uCost = (double)dImgVal;
            uCost /= 255.0;

            if(c == start_col)
                if(r != start_row)
                    uCost = 999999;

            // Compute binary cost, to get from column c-1 to c
            const int MAX_STEP = 2;
            const double STEP_COST = 0.75;

            for(int dr = -MAX_STEP; dr <= MAX_STEP; ++dr)
            {
                int back_r = r + dr;
                if(back_r < 0 || back_r >= dImg.rows)
                    continue;

                double binaryCost = 0;
                if(dr != 0)
                    binaryCost = STEP_COST;

                double backCost = 0;
                if(c > 0)
                    backCost = costMatrix.at<double>(back_r, c-1);

                double tentativeCost = backCost + binaryCost + uCost;
                if(tentativeCost < bestCost)
                {
                    bestCost = tentativeCost;
                    bestBack = dr;
                }
            }

            costMatrix.at<double>(r,c) = bestCost;
            backMatrix.at<int>(r,c) = bestBack;
        }
    }


    Mat endCost = costMatrix.col(costMatrix.cols-1);
    double minval = 0;
    double maxval = 0;
    Point minpt;
    Point maxpt;

    minMaxLoc(endCost,&minval,&maxval,&minpt,&maxpt);
    endCost.at<double>(minpt.y,minpt.x) = 999999;
    int minRow = minpt.y;

    const int STEP_DIST = 40;
    for(int c = backMatrix.cols - 1; c > start_col; --c)
    {
        int dr = backMatrix.at<int>(minRow,c);
        minRow += dr;
        if(c % STEP_DIST == 0)
            points.push_back(QPoint(c,minRow));
    }

    std::reverse(points.begin(), points.end());
    std::vector<QPoint> points_smooth = smoothBaseline(points, 10, -1, -1);

    for(unsigned int i = 0; i < points_smooth.size(); ++i)
    {
        int px = points_smooth[i].x();
        int py = points_smooth[i].y();
        double winX, winY;

        double widthScale = (double)window_width/(double)imW;
        double heightScale = (double)window_height/(double)imH;
        winX = px * widthScale;
        winY = py * heightScale;
        winY = window_height - winY;

        int k = -1;
        double u = -1;
        double v = -1;
        Vector3d newpoint;

        clickOnMesh(winX,winY,&k,&u,&v,&newpoint);

        if(k >= 0)
        {
            mesh->floatingLineConstraint.addSuggestionPoint(k,u,v);
        }
    }
}

QString meshname = "foo";
int numInterpFrames = 0;
int interpCount = 0;

void FlatGLWidget::startInterpolation()
{
    if(interpolation_t >= 0.99)
    {
        interpolation_t = 0.0;
        updateGL();
    }
    else
    {
        interpolation_t = 0.0;
        interpCount = 0;
        if(interpTimer)
        {
            interpTimer->stop();
            delete interpTimer;
        }

        numInterpFrames = QInputDialog::getInt(0, tr("Interp Frames"), tr("Select number of frames"));

        interpTimer = new QTimer();
        connect(interpTimer, SIGNAL(timeout()), this, SLOT(interpolationStep()));
        interpTimer->start(10);
        renderInterpolation = true;
    }
}

void FlatGLWidget::interpolationStep()
{
    double t_step = 1.0 / numInterpFrames;
    interpolation_t = interpolation_t + t_step;
    if(interpolation_t >= 1.0)
    {
        interpTimer->stop();
    }
    if(SPIN_AROUND)
    {
        rot_angle_new = rot_angle_new + (2*3.1415) / (1.0/t_step);
    }

    updateGL();
    /*
    renderEnhanced = false;
    updateGL();
    QPixmap stepPixmap = renderPixmap(window_width,window_height,true);
    //QString pixmapFilename = ParchmentFlattener::meshDirectory;
    //pixmapFilename.append(meshname).append("_").append(QString::number(interpCount)).append(".png");
    //stepPixmap.save(pixmapFilename);

    renderEnhanced = true;
    updateGL();
    QPixmap stepPixmap2 = renderPixmap(window_width,window_height,true);
    //QString pixmapFilename2 = ParchmentFlattener::meshDirectory;
    //pixmapFilename2.append("wire").append("_").append(QString::number(interpCount)).append(".png");
    //stepPixmap2.save(pixmapFilename2);

    QImage imShade = stepPixmap.toImage();
    QImage imEnhanced = stepPixmap2.toImage();
    QImage imBlended;
    imBlended = QImage(imShade.size(), QImage::Format_ARGB32_Premultiplied);

    QPainter painter(&imBlended);
    painter.setOpacity(1.0);
    painter.drawImage(0,0,imShade);
    painter.setOpacity(interpolation_t);
    painter.drawImage(0,0,imEnhanced);
    painter.end();

    QString blendFilename = ParchmentFlattener::meshDirectory;
    blendFilename.append(meshname).append("_").append(QString::number(interpCount)).append(".png");
    imBlended.save(blendFilename);
    */

    ++interpCount;
}

void FlatGLWidget::setModelViewMatrixGL()
{
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    if(renderInterpolation)
    {
        Vector3d disp = Vector3d(qSin(rot_angle_new), qCos(rot_angle_new), camz_interp);
        disp.normalize();
        Vector3d eye = target_new - camz * disp;
        gluLookAt(eye(0),eye(1),eye(2),target_new(0),target_new(1),target_new(2),0,0,1);
    }
    else
    {
        Vector3d eye = target_new - camz * currentGravity;
        //gluLookAt(eye(0),eye(1),eye(2),target_new(0),target_new(1),target_new(2),currentUpVec(0),currentUpVec(1),currentUpVec(2));
        QMatrix4x4 mview_matrix;
        QVector3D qEye(eye(0),eye(1),eye(2));
        QVector3D qTarget(target_new(0),target_new(1),target_new(2));
        QVector3D qUp(currentUpVec(0),currentUpVec(1),currentUpVec(2));
        mview_matrix.setToIdentity();
        mview_matrix.lookAt(qEye,qTarget,qUp);
        glMultMatrixd(mview_matrix.constData());
    }
}

void FlatGLWidget::setModelViewMatrixGL(Vector3d mvTarget)
{
    Vector3d eye = mvTarget - camz * currentGravity;
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(eye(0),eye(1),eye(2),mvTarget(0),mvTarget(1),mvTarget(2),currentUpVec(0),currentUpVec(1),currentUpVec(2));
}

int FlatGLWidget::selectLineConstraintKeypoint(int mouseX, int mouseY)
{
    int winX = mouseX;
    int winY = window_height - mouseY;
    QPointF pm((qreal)winX, (qreal)winY);

    LineConstraint lc = mesh->lineConstraints[mesh->selectedLineConstraint];
    for(int i = 0; i < lc.size(); ++i)
    {
        Vector3d sp = getSurfacePoint(lc.faces[i], lc.sts(0,i), lc.sts(1,i));
        QPointF pv = fullProjectionGL(sp(0), sp(1), sp(2));

        double dist = QLineF(pm,pv).length();

        if(dist <= CLICK_THRESH)
        {
            return i;
        }
    }
    return -1;
}

void FlatGLWidget::setHoverPoint(int winX, int winY)
{
    int winX2 = winX;
    int winY2 = window_height - winY;

    int k = 0;
    double u = 0;
    double v = 0;
    Vector3d p;
    bool isIntersect = clickOnMesh(winX2,winY2,&k,&u,&v,&p);
    mesh->isHover = isIntersect;
    if(isIntersect)
    {
        mesh->hover_k = k;
        mesh->hover_u = u;
        mesh->hover_v = v;
        emit hoverChangedSignal();
        emit hoverChanged(winX2,winY2);
    }
}

QPointF FlatGLWidget::fullProjectionGL(double x, double y, double z)
{
    setProjectionMatrixGL();
    setModelViewMatrixGL();

    GLdouble modelview[16], projection[16];
    GLint viewport[4];

    glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
    glGetDoublev(GL_PROJECTION_MATRIX, projection);
    glGetIntegerv(GL_VIEWPORT, viewport);

    double tx, ty, tz;

    gluProject(x, y, z, modelview, projection, viewport, &tx, &ty, &tz);

    return QPointF(tx,ty);
}

Vector3d FlatGLWidget::fullUnProjectionGL(double winx, double winy, double winz)
{
    setProjectionMatrixGL();
    setModelViewMatrixGL();

    GLdouble modelview[16], projection[16];
    GLint viewport[4];

    glGetDoublev(GL_MODELVIEW_MATRIX, modelview);
    glGetDoublev(GL_PROJECTION_MATRIX, projection);
    glGetIntegerv(GL_VIEWPORT, viewport);

    double tx, ty, tz;

    gluUnProject(winx, winy, winz, modelview, projection, viewport, &tx, &ty, &tz);

    return Vector3d(tx,ty,tz);
}

void FlatGLWidget::setProjectionMatrixGL()
{
    double view_ratio = (double)window_width/(double)window_height;
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    QMatrix4x4 projMatrix;
    projMatrix.setToIdentity();
    projMatrix.perspective(45.0,view_ratio,znear,zfar);
    glMultMatrixd(projMatrix.constData());
    //gluPerspective(45.0,view_ratio,znear,zfar);   
}

void FlatGLWidget::zoomIn()
{
    if(smoothZoom)
    {
        camz += DELTA_ZOOM;
    }
    else
    {
        camz *= 0.8;
    }
    qDebug() << camz;
    updateGL();
}

void FlatGLWidget::zoomOut()
{
    if(smoothZoom)
    {
        camz -= DELTA_ZOOM;
    }
    else
    {
        camz /= 0.8;
    }
    updateGL();
}

Vector3d FlatGLWidget::getTarget()
{
    return target_new;
}

void FlatGLWidget::setTarget(Vector3d _target)
{
    target = _target;
    target_new = _target;
    smoothZoom = true;
}

void FlatGLWidget::removeBlackBorderBySize()
{
    double aveArea = 0;
    for(int k = 0; k < mesh->F.cols(); ++k)
    {
        int i0 = mesh->F(0,k);
        int i1 = mesh->F(1,k);
        int i2 = mesh->F(2,k);

        Vector3d v0 = mesh->V.col(i0);
        Vector3d v1 = mesh->V.col(i1);
        Vector3d v2 = mesh->V.col(i2);

        Vector3d v01 = v1-v0;
        Vector3d v02 = v2-v0;

        Vector3d cross = v01.cross(v02);
        double area = cross.norm() / 2.0;
        aveArea += area;
    }
    aveArea /= mesh->F.cols();
    double thresh = aveArea * 5;
    QVector<int> badFaces;
    for(int k = 0; k < mesh->F.cols(); ++k)
    {
        int i0 = mesh->F(0,k);
        int i1 = mesh->F(1,k);
        int i2 = mesh->F(2,k);

        Vector3d v0 = mesh->V.col(i0);
        Vector3d v1 = mesh->V.col(i1);
        Vector3d v2 = mesh->V.col(i2);

        Vector3d v01 = v1-v0;
        Vector3d v02 = v2-v0;

        Vector3d cross = v01.cross(v02);
        double area = cross.norm() / 2.0;
        if(area > thresh)
            badFaces.push_back(k);
    }

    for(int i = 0; i < badFaces.size(); ++i)
    {
        mesh->removeFace(badFaces[i] - i);
    }
}

void FlatGLWidget::removeBlackBorder()
{
    glColor3d(1,1,1);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_DEPTH_TEST);
    int currentTex = -1;

    QGLPixelBuffer *pixBuffer = new QGLPixelBuffer(128, 128, QGLFormat::defaultFormat(), this);
    pixBuffer->makeCurrent();
    glClearColor(0,0,0,1);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluOrtho2D(-1,1,-1,1);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    for(int k = 0; k < mesh->F.cols(); k += 500)
    {
        int uv0 = mesh->FTC(k,0);
        int uv1 = mesh->FTC(k,1);
        int uv2 = mesh->FTC(k,2);

        double u0 = mesh->TC(uv0,0);    double v0 = mesh->TC(uv0,1);
        double u1 = mesh->TC(uv1,0);    double v1 = mesh->TC(uv1,1);
        double u2 = mesh->TC(uv2,0);    double v2 = mesh->TC(uv2,1);

        int m = mesh->Materials(k);
        if(m != currentTex)
        {
            qDebug() << "Bind Texture " << m;
            currentTex = m;
            glBindTexture(GL_TEXTURE_2D, meshTextures[currentTex]);
        }

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        glBegin(GL_TRIANGLES);
        glTexCoord2d(u0,v0);
        glVertex2d(-1,-1);

        glTexCoord2d(u1,v1);
        glVertex2d(-1,1);

        glTexCoord2d(u2,v2);
        glVertex2d(1,1);
        glEnd();

//        glBegin(GL_TRIANGLES);
//        glTexCoord2d(mesh->TC(uv0,0), mesh->TC(uv0,1));
//        glVertex3d(-1,-1,0);

//        glTexCoord2d(mesh->TC(uv1,0), mesh->TC(uv1,1));
//        glVertex3d(1,-1,0);

//        glTexCoord2d(mesh->TC(uv2,0), mesh->TC(uv2,1));
//        glVertex3d(1,1,0);
//        glEnd();

        QString imName = QString("im").append(QString::number(k));

        QImage im = pixBuffer->toImage();
        QString filename = ParchmentFlattener::meshDirectory;
        filename.append(imName);
        filename.append(".jpg");
        im.save(filename);


    }
    delete pixBuffer;
    makeCurrent();
    setProjectionMatrixGL();
    setModelViewMatrixGL();
}
