//
//  FlatteningEngine.cpp
//  ParchmentFlattener
//
//  Copyright (c) 2014 Kazim Pal.
//  Released under the MIT License.
//
#define IGL_HEADER_ONLY
#define PI 3.141592654
#include "ParchmentFlattener.h"
#include "FlatteningEngine.h"
#include "IGLMesh.h"
#include "igl/is_border_vertex.h"
#include "igl/adjacency_list.h"
#include <QDebug>
#include <vector>
#include "lim/src/LIMSolverInterface.h"
#include <Eigen/Sparse>
#include <Eigen/SparseLU>
#include <QFile>

using namespace std;

double prev_diff = 0;

FlatteningEngine::FlatteningEngine(void)
{
    limData = 0;
}

FlatteningEngine::~FlatteningEngine(void)
{
    deleteLimData();
}

MatrixXd FlatteningEngine::flattenTutte(IGLMesh *wMesh)
{
    MatrixXd V = wMesh->V;
    MatrixXi F = wMesh->F;

    MatrixXd vertices = V.transpose();
    MatrixXi faces = F.transpose();

    qDebug() << "Computing Tutte Border";

    vector<int> borderVertices;
    vector<bool> isBorder = igl::is_border_vertex(vertices, faces);
    vector<bool> visited;

    for(unsigned int i = 0; i < isBorder.size(); ++i)
    {
        if(isBorder[i] && borderVertices.size() == 0)
        {
            borderVertices.push_back(i);
            visited.push_back(true);
        }
        else
        {
            visited.push_back(false);
        }
    }

    vector< vector<int> > A;
    igl::adjacency_list(faces, A, true);

    bool finished = false;
    while(!finished)
    {
        finished = true;
        int currentV = borderVertices.back();
        for(unsigned int i = 0; i < A[currentV].size(); ++i)
        {
            int n = A[currentV][i];
            if(isBorder[n])
            {
                if(visited[n] == false)
                {
                    finished = false;
                    borderVertices.push_back(n);
                    visited[n] = true;
                    break;
                }
            }
        }
    }

    Vector3d C = wMesh->C;
    double maxLength = 0;
    for(unsigned int i = 0; i < borderVertices.size(); ++i)
    {
        int b_index = borderVertices[i];
        Vector3d pi = vertices.row(b_index);
        Vector3d di = pi - C;
        di(2) = 0;
        double length = di.norm();
        if(length > maxLength)
        {
            maxLength = length;
        }
    }
    maxLength *= 1.1;

    double dTheta = 360.0 / borderVertices.size();
    for(unsigned int i = 0; i < borderVertices.size(); ++i)
    {
        int b_index = borderVertices[i];
        double theta_i = dTheta * i * PI / 180.0;
        double x = maxLength * cos(theta_i);
        double y = maxLength * sin(theta_i);
        Vector3d pi(x,y,0);
        vertices.row(b_index) = pi;
    }

    qDebug() << "Computing Tutte Mapping";

    int numRows = V.cols();
    SparseMatrix<double, ColMajor> Ax(numRows,numRows);
    SparseMatrix<double, ColMajor> Ay(numRows,numRows);
    VectorXd bx, by;
    VectorXd X, Y;
    bx.resize(numRows,1);
    by.resize(numRows,1);
    MatrixXd Vflat = vertices.transpose();

    typedef Eigen::Triplet<double> T;
    std::vector<T> tripletList;

    for(int i = 0; i < V.cols(); ++i)
    {
        if(isBorder[i])
        {
            double xi = vertices(i,0);
            double yi = vertices(i,1);
            tripletList.push_back(T(i,i,1));
            bx(i,0) = xi;
            by(i,0) = yi;
        }
        else
        {

            int valence = A[i].size();
            tripletList.push_back(T(i,i,1));
            double w = -1.0/valence;
            for(int j = 0; j < valence; ++j)
            {
                int ni = A[i][j];
                tripletList.push_back(T(i,ni,w));
            }
            bx(i,0) = 0;
            by(i,0) = 0;
        }
    }

    Ax.setFromTriplets(tripletList.begin(), tripletList.end());
    Ay.setFromTriplets(tripletList.begin(), tripletList.end());

    qDebug() << "Solving Tutte X Coordinates";
    SparseLU< SparseMatrix< double , ColMajor >, NaturalOrdering<int> > solver;
    solver.analyzePattern(Ax);
    solver.factorize(Ax);
    X = solver.solve(bx);

    qDebug() << "Solving Tutte Y Coordinates";
    solver.analyzePattern(Ay);
    solver.factorize(Ay);
    Y = solver.solve(by);

    vertices.col(0) = X;
    vertices.col(1) = Y;

    Vflat = vertices.transpose();
    qDebug() << "Saving tutte vertices";

    Vflat.row(2).setZero();
    wMesh->setFlatVertexPositions(Vflat);
    QString tutteFilename = ParchmentFlattener::meshDirectory;
    tutteFilename.append("tutte.obj");
    wMesh->saveFlatMesh(tutteFilename);
    return Vflat;
}

int FlatteningEngine::initialFlattenTuttePlusLIM(IGLMesh *wMesh)
{
    qDebug() << "Initializing with Tutte's Mapping";
    MatrixXd V_Tutte = flattenTutte(wMesh);
    QVector<int> inv = wMesh->containsInvertedElements(V_Tutte);
    if(inv.size() > 0)
    {
        qDebug() << "contains inverted elements";
        return -2;
    }

    for(int i = 0; i < inv.size(); ++i)
    {
        //wMesh->removeFace(inv[i]);
    }
    wMesh->containsInvertedElements(V_Tutte);
    MatrixXd Vinit = wMesh->V;
    MatrixXi F = wMesh->F;

    Matrix<double,Dynamic,3> vertices = V_Tutte.transpose();
    Matrix<double,Dynamic,3> initialVertices = Vinit.transpose();
    Matrix<int,Dynamic,3> faces = F.transpose();

    MatrixXd V_Tutte_t = V_Tutte.transpose();
    MatrixXi F_t = F.transpose();

    vector<int> borderVertices;
    std::vector<bool> isBorder = igl::is_border_vertex(V_Tutte_t,F_t);
    for(unsigned int i = 0; i < isBorder.size(); ++i)
    {
        if(isBorder[i])
            borderVertices.push_back(i);
    }

    SparseMatrix<double> constraintMatrix;
    Matrix<double,Dynamic,1> constraintTargets;
    Matrix<double,Dynamic,1> gradients;

    constraintMatrix.resize(2,Vinit.cols()*2);
    constraintMatrix.coeffRef(0,0) = 1;
    constraintMatrix.coeffRef(1,1) = 1;

    constraintTargets.resize(2,1);
    constraintTargets(0,0) = V_Tutte(0,0);
    constraintTargets(1,0) = V_Tutte(1,0);

    qDebug() << "Solving ARAP From Tutte Initialization";
    int result = -2;
    result = ComputeLIM(vertices,initialVertices,faces,borderVertices,gradients,constraintMatrix,constraintTargets,3,0.00001,10000,true);

    qDebug() << "LIM Result: " << result;
    if(result >= -1)
    {
        qDebug() << "Setting flat vertices";
        MatrixXd Vflat = vertices.transpose();
        Vflat.row(2).setZero();
        wMesh->setFlatVertexPositions(Vflat);
        wMesh->setInitialVertexPositions(Vflat);
        wMesh->flattened = true;
        wMesh->reGatherCurrentConstraints();
    }
    return result;
}

void FlatteningEngine::rebuildLimData(IGLMesh *wMesh)
{
    qDebug() << "rebuilding limdata";
    MatrixXd V = wMesh->flatV;
    MatrixXd Vinit = wMesh->Vinitial;
    MatrixXd Vorig = wMesh->V;
    MatrixXi F = wMesh->F;

    vector<int> borderVertices;

    double len_avex = 0;
    double len_avey = 0;
    for(int i = 0; i < F.cols(); ++i)
    {
        double lenx = wMesh->perFaceScaleF(0,i);
        double leny = wMesh->perFaceScaleF(1,i);
        len_avex += lenx;
        len_avey += leny;
    }
    len_avex /= F.cols();
    len_avey /= F.cols();

    Matrix<double,Dynamic,1> gradients;
    gradients.setZero(F.cols() * 4, 1);
    for(int i = 0; i < F.cols(); ++i)
    {
        int xxi = i * 2;
        int xyi = i * 2 + 1;
        int yxi = (F.cols() + i) * 2;
        int yyi = (F.cols() + i) * 2 + 1;

        Vector3d ri = wMesh->perFaceRightVectorsF.col(i);
        Vector3d ui = wMesh->perFaceUpVectorsF.col(i);
        double lenx = wMesh->perFaceScaleF(0,i);
        double leny = wMesh->perFaceScaleF(1,i);

        Vector2d R(ri(0),ri(1));
        Vector2d U(ui(0),ui(1));
        MatrixXd basis;
        basis.setZero(2,2);
        basis.col(0) = R;
        basis.col(1) = U;

        basis.col(0) = basis.col(0) * lenx;
        basis.col(1) = basis.col(1) * leny;

        Vector2d gradX = Vector2d(1,0);
        Vector2d gradY = Vector2d(0,1);

        MatrixXd basis_I = basis.inverse();
        double len_target = 0.1;
        if(wMesh->scaleConstraints2.size() > 0)
            len_target = wMesh->scaleConstraints2[0].scale(1);
        basis_I = basis_I * len_target;

        Vector2d gradX_T = basis_I * gradX;
        Vector2d gradY_T = basis_I * gradY;

        gradients(xxi) = gradX_T(0);
        gradients(xyi) = gradX_T(1);
        gradients(yxi) = gradY_T(0);
        gradients(yyi) = gradY_T(1);
    }

    SparseMatrix<double> constraintMatrix;
    Matrix<double,Dynamic,1> constraintTargets;

    if(wMesh->lineConstraints.size() > 0)
    {
        int totalConstraints = 0;
        for(int k = 0; k < wMesh->lineConstraints.size(); ++k)
        {
            LineConstraint lc = wMesh->lineConstraints[k];
            totalConstraints += lc.size() - 1;
        }
        constraintMatrix.resize(totalConstraints,V.cols()*2);
        constraintTargets.setZero(totalConstraints,1);

        int constraintCount = 0;
        for(int k = 0; k < wMesh->lineConstraints.size(); ++k)
        {
            LineConstraint lc = wMesh->lineConstraints[k];

            if(!lc.isActive())
                continue;

            for(int i = 0; i < lc.size() - 1; ++i)
            {
                int fi = lc.faces[i];
                double ui = lc.sts(0,i);
                double vi = lc.sts(1,i);
                double wi = 1 - ui - vi;
                int v0 = wMesh->F(0,fi);
                int v1 = wMesh->F(1,fi);
                int v2 = wMesh->F(2,fi);

                int v0x = v0*2;
                int v1x = v1*2;
                int v2x = v2*2;

                int v0y = v0*2+1;
                int v1y = v1*2+1;
                int v2y = v2*2+1;

                if(lc.mIsVertical)
                {
                    constraintMatrix.coeffRef(constraintCount,v0x) += wi;
                    constraintMatrix.coeffRef(constraintCount,v1x) += ui;
                    constraintMatrix.coeffRef(constraintCount,v2x) += vi;
                }
                else
                {
                    constraintMatrix.coeffRef(constraintCount,v0y) += wi;
                    constraintMatrix.coeffRef(constraintCount,v1y) += ui;
                    constraintMatrix.coeffRef(constraintCount,v2y) += vi;
                }

                fi = lc.faces[i+1];
                ui = lc.sts(0,i+1);
                vi = lc.sts(1,i+1);
                wi = 1 - ui - vi;
                v0 = wMesh->F(0,fi);
                v1 = wMesh->F(1,fi);
                v2 = wMesh->F(2,fi);

                v0x = v0*2;
                v1x = v1*2;
                v2x = v2*2;

                v0y = v0*2+1;
                v1y = v1*2+1;
                v2y = v2*2+1;

                if(lc.mIsVertical)
                {
                    constraintMatrix.coeffRef(constraintCount,v0x) -= wi;
                    constraintMatrix.coeffRef(constraintCount,v1x) -= ui;
                    constraintMatrix.coeffRef(constraintCount,v2x) -= vi;
                }
                else
                {
                    constraintMatrix.coeffRef(constraintCount,v0y) -= wi;
                    constraintMatrix.coeffRef(constraintCount,v1y) -= ui;
                    constraintMatrix.coeffRef(constraintCount,v2y) -= vi;
                }

                constraintTargets(constraintCount,0) = 0;

                ++constraintCount;
            }
        }
    }
    else
    {
        constraintMatrix.resize(2,V.cols()*2);
        constraintMatrix.insert(0,0) = 1;
        constraintMatrix.insert(1,1) = 1;
        constraintTargets.setZero(2,1);
        double x0 = Vinit(0,0);
        double y0 = Vinit(1,0);
        constraintTargets(0,0) = x0;
        constraintTargets(1,0) = y0;
    }

    deleteLimData();

    Matrix<double,Dynamic,3> vertices = Vinit.transpose();
    Matrix<double,Dynamic,3> initialVertices = Vorig.transpose();
    Matrix<int,Dynamic,3> faces = F.transpose();

    limData = InitLIM(vertices,initialVertices,faces,borderVertices,gradients,constraintMatrix,constraintTargets,5,0.00001,1000,true);
    limData->solver->EnableBarriers = true;

    prev_diff = 0;
}

int FlatteningEngine::singleStep(IGLMesh *wMesh)
{
    MatrixXd V = wMesh->flatV;
    Matrix<double,Dynamic,3> vertices;
    vertices.setZero(V.cols(), 3);

    /*
    if(limData->solver->EnableBarriers)
        qDebug() << "Barriers are enabled";
    else
        qDebug() << "Barriers are disabled";
    */
    int result = ComputeLIM_Step(limData,vertices);
    int finished = -1;
    if(result > -1)
    {
        MatrixXd Vnew = vertices.transpose();
        Vnew.row(2).setZero();
        wMesh->setFlatVertexPositions(Vnew);
        wMesh->flattened = true;

        MatrixXd diffV = wMesh->flatV - V;
        double total_diff = diffV.array().square().sum();
        double d2red = qAbs(total_diff - prev_diff);
        qDebug() << "Step Diff Sum: " << total_diff << " Diff Diff: " << d2red;

        if(total_diff < 0.1 || total_diff != total_diff)
            finished = 1;
        else
            prev_diff = total_diff;
    }
    return finished;
}

void FlatteningEngine::deleteLimData()
{
    if(limData)
    {
        FreeLIMData(limData);
        delete limData;
    }
}
