//
//  ParchmentFlattener.cpp
//  ParchmentFlattener
//
//  Copyright (c) 2014 Kazim Pal.
//  Released under the MIT License.
//
#include "ParchmentFlattener.h"
#include "FlatteningEngine.h"
#include "IGLMesh.h"
#include <QKeyEvent>
#include <qtimer.h>
#include <QtConcurrentRun>
#include <QFileDialog>
#include <QInputDialog>
#include <QMessageBox>

#define INIT_ALL false

ParchmentFlattener::MarkupMode ParchmentFlattener::markupMode = ParchmentFlattener::None;
bool ParchmentFlattener::verticalLines = false;
QString ParchmentFlattener::meshFilename = "foo";
QString ParchmentFlattener::meshDirectory = "foo";


ParchmentFlattener::ParchmentFlattener(QWidget *parent, Qt::WFlags flags)	: QMainWindow(parent, flags)
{
	ui.setupUi(this);
    mesh = new IGLMesh();
    fGlWidget = 0;

    if(INIT_ALL)
    {
        qDebug() << "INITIALIZING ALL MESHES";
        initAll();
    }
    else
    {
        loadMesh();
        if(mesh == 0)
        {
            QApplication::quit();
        }
    }


    fGlWidget = new FlatGLWidget(mesh, this);

    ui.rightFrame->layout()->addWidget(fGlWidget);

    markupModeGroup = new QActionGroup(this);
    markupModeGroup->addAction(ui.actionOrientation);
    markupModeGroup->addAction(ui.actionScale);

    connect(ui.actionFlatten, SIGNAL(triggered()), this, SLOT(flatten()));
    connect(ui.actionOrientation, SIGNAL(triggered()), this, SLOT(toggleLineMode()));
    connect(ui.actionScale, SIGNAL(triggered()), this, SLOT(toggleScaleMode()));
    connect(ui.actionVertical,SIGNAL(triggered()), this, SLOT(toggleVertical()));
    connect(ui.actionLoad_Page, SIGNAL(triggered()), this, SLOT(loadMesh()));
    connect(ui.actionSave,SIGNAL(triggered()),this,SLOT(saveConstraintFile()));
    connect(ui.actionLoad,SIGNAL(triggered()), this, SLOT(loadConstraintFile()));
    connect(ui.actionClear,SIGNAL(triggered()), this, SLOT(clearConstraints()));
    connect(ui.actionSave_Flat,SIGNAL(triggered()),this,SLOT(saveFlatMeshFile()));
    connect(ui.actionNext,SIGNAL(triggered()),this,SLOT(nextConstraint()));
    connect(ui.actionPrevious,SIGNAL(triggered()),this,SLOT(prevConstraint()));
    connect(ui.actionDelete,SIGNAL(triggered()),this,SLOT(deleteConstraint()));
    connect(ui.actionCrossField,SIGNAL(triggered()),this,SLOT(toggleRenderCrossField()));

    connect(ui.actionStandard,SIGNAL(triggered()),this,SLOT(setRenderModeStandard()));
    connect(ui.actionStretch,SIGNAL(triggered()),this,SLOT(setRenderModeStretch()));

    connect(ui.actionSnapshot,SIGNAL(triggered()),this,SLOT(snapshot()));

    connect(mesh,SIGNAL(smoothingIterationFinished()),fGlWidget,SLOT(smoothingIterationWasDone()));

    connect(mesh,SIGNAL(flatteningIterationFinished()),fGlWidget,SLOT(flatteningIterationWasDone()));

    connect(mesh,SIGNAL(autoSave()),this,SLOT(autosaveConstraintFile()));

    fGlWidget->loadTextures();

    mesh->startSmoothing();
}

ParchmentFlattener::~ParchmentFlattener()
{

}

void ParchmentFlattener::initAll()
{
    QString root = "/host/Users/Kazim/Pictures/LMA/ToFlattenPoisson/";
    QDir root_dir(root);
    root_dir.setFilter(QDir::Dirs);
    QStringList sub_dirs = root_dir.entryList();

    QString logFileName = root;
    logFileName.append("initall_log.txt");
    QFile logfile(logFileName);
    logfile.open(QIODevice::WriteOnly | QIODevice::Text);
    QTextStream logFileOut(&logfile);

    //QString failLogFileName = root;
    //failLogFileName.append("fail_log.txt");
    //QFile failLogfile(logFileName);
    //failLogfile.open(QIODevice::WriteOnly | QIODevice::Text);
    //QTextStream failLogFileOut(&failLogfile);

    sub_dirs.removeOne(".");
    sub_dirs.removeOne("..");

    for(int i = 0; i < sub_dirs.size(); ++i)
    {
        delete mesh;
        mesh = new IGLMesh();
        QString sub_dir = sub_dirs.at(i);
        qDebug() << "Processing page " << sub_dir;
        QString cleanFileName = root;
        cleanFileName.append(sub_dir).append("/result.nvm.cmvs/00/visualize/texturedMeshNew.obj");
        QFile F_clean(cleanFileName);
        if(!F_clean.exists())
        {
            logFileOut << sub_dir << ": couldn't find cleanfile.\n";
            //failLogFileOut << sub_dir << ": couldn't find cleanfile.\n";
            continue;
        }

        QString flatInitFileName = root;
        flatInitFileName.append(sub_dir).append("/result.nvm.cmvs/00/visualize/texturedMeshNewFlatInit.obj");
        QFile F_init(flatInitFileName);
        if(F_init.exists())
        {
            //logFileOut << sub_dir << ": skipped. already initialized.\n";
            qDebug() << "already init";
            continue;
        }

        ParchmentFlattener::meshFilename = cleanFileName;
        mesh->loadMesh(ParchmentFlattener::meshFilename);

        if(mesh->containsUnreferecedVertices())
        {
            logFileOut << sub_dir << ": skipped. unreferenced vertices.\n";
            //failLogFileOut << sub_dir << ": skipped. unreferenced vertices.\n";
            qDebug() << "unreferenced vertices";
            continue;
        }

        if(mesh->containsStrayComponents())
        {
            logFileOut << sub_dir << ": skipped. stray components.\n";
            //failLogFileOut << sub_dir << ": skipped. stray components.\n";
            qDebug() << "stray components";
            continue;
        }

        ParchmentFlattener::meshDirectory = ParchmentFlattener::meshFilename;
        int finalslashindex = ParchmentFlattener::meshDirectory.lastIndexOf("/");
        ParchmentFlattener::meshDirectory.chop(ParchmentFlattener::meshFilename.length() - finalslashindex - 1);

        int result = mesh->flatteningEngine->initialFlattenTuttePlusLIM(mesh);
        if(result > 0)
        {
            mesh->saveFlatVertices(flatInitFileName);
            logFileOut << sub_dir << ": success.\n";
        }
        else
        {
            logFileOut << sub_dir << ": flatten failed. result " << result << "\n";
            //failLogFileOut << sub_dir << ": flatten failed. result " << result << "\n";
        }
    }

    logfile.close();
    //failLogfile.close();
}

void ParchmentFlattener::setMarkupMode(MarkupMode m)
{
    if(m != markupMode)
    {
        markupMode = m;
    }
    switch(m)
    {
    default:
        //fGlWidget->setMouseTracking(false);
        break;
    }

    fGlWidget->updateGL();
}

void ParchmentFlattener::setNoneMode()
{
    setMarkupMode(ParchmentFlattener::None);
    ui.actionOrientation->setChecked(false);
    ui.actionScale->setChecked(false);

    ui.actionVertical->setEnabled(false);
    setVertical(false);

    fGlWidget->setMouseTracking(false);
}

void ParchmentFlattener::setLineMode()
{
    setMarkupMode(ParchmentFlattener::Line);
    ui.actionOrientation->setChecked(true);

    ui.actionVertical->setEnabled(true);
    setVertical(false);

    //fGlWidget->setMouseTracking(false);
    fGlWidget->setMouseTracking(true);
}

void ParchmentFlattener::setScaleMode()
{
    setMarkupMode(ParchmentFlattener::Scale);
    ui.actionScale->setChecked(true);

    ui.actionVertical->setEnabled(false);
    setVertical(false);

    fGlWidget->setMouseTracking(true);
}

void ParchmentFlattener::toggleVertical()
{
    setVertical(!verticalLines);
}

void ParchmentFlattener::setVertical(bool isVertical)
{
    verticalLines = isVertical;
    if(verticalLines)
    {
        ui.actionVertical->setIcon(QIcon(":/icons/verticon.png"));
        //ui.actionVertical->setChecked(true);
    }
    else
    {
        ui.actionVertical->setIcon(QIcon(":/icons/horizicon.png"));
        //ui.actionVertical->setChecked(false);
    }
}

void ParchmentFlattener::loadMesh()
{
    QString newFileName = QFileDialog::getOpenFileName(0,tr("Load Parchment"),"/host/Users/Kazim/Pictures/LMA/ToFlattenPoisson/",tr("Obj files (*.obj)"));

    QFile F(newFileName);
    if(!F.exists())
    {
        delete mesh;
        mesh = 0;
        return;
    }

    mesh->disallowSmoothing();
    while(!mesh->smoothIsPaused)
    {
        qDebug() << "Waiting for smoothing iteration to finish";
    }

    ParchmentFlattener::meshFilename = newFileName;
    mesh->loadMesh(ParchmentFlattener::meshFilename);

    ParchmentFlattener::meshDirectory = ParchmentFlattener::meshFilename;
    int finalslashindex = ParchmentFlattener::meshDirectory.lastIndexOf("/");
    ParchmentFlattener::meshDirectory.chop(ParchmentFlattener::meshFilename.length() - finalslashindex - 1);

    QString flatFilename = ParchmentFlattener::meshFilename;
    flatFilename.chop(4);
    flatFilename.append("FlatInit.obj");
    if(QFile(flatFilename).exists())
    {
        qDebug() << "Loading initial configuration from " << flatFilename;
        mesh->loadFlatVertices(flatFilename);
    }
    else
    {
        qDebug() << "Computing initial configuration";
        mesh->flatteningEngine->initialFlattenTuttePlusLIM(mesh);
        mesh->saveFlatVertices(flatFilename);
    }
    qDebug() << "finished initialzing. now rebuilding limdata";
    mesh->flatteningEngine->rebuildLimData(mesh);
    if(fGlWidget)
    {
        fGlWidget->loadTextures();
    }

    mesh->allowSmoothing();
}

void ParchmentFlattener::flatten()
{
    mesh->flatteningLocked = !mesh->flatteningLocked;
    if(mesh->flatteningLocked)
    {
        ui.actionFlatten->setChecked(false);
        mesh->flatteningEnabled = false;
    }
    else
    {
        ui.actionFlatten->setChecked(true);
        fGlWidget->resetRotation();
        mesh->flatteningEnabled = true;
    }
}

void ParchmentFlattener::clearConstraints()
{
    mesh->clearConstraints();
}

void ParchmentFlattener::keyPressEvent(QKeyEvent *e)
{
    if (e->key() == Qt::Key_S)
    {
        toggleScaleMode();
    }
    else if(e->key() == Qt::Key_L)
    {
        toggleLineMode();
    }
    else if(e->key() == Qt::Key_Tab)
    {
        nextConstraint();
    }
    else if(e->key() == Qt::Key_Delete)
    {
        deleteConstraint();
    }
    else if(e->key() == Qt::Key_C)
    {
        toggleRenderCrossField();
    }
    else if(e->key() == Qt::Key_Control)
    {
        //fGlWidget->renderOriginal = !fGlWidget->renderOriginal;
        //fGlWidget->updateGL();
    }
    else if(e->key() == Qt::Key_P)
    {
        mesh->floatingLineConstraint.acceptSuggestions();
        fGlWidget->generateSuggestionTailDynamic();
        fGlWidget->updateGL();
    }
    else if(e->key() == Qt::Key_I)
    {
        //fGlWidget->renderInterpolation = !fGlWidget->renderInterpolation;
        //fGlWidget->updateGL();
    }
    else if(e->key() == Qt::Key_Semicolon)
    {
        if(fGlWidget->renderConstraints)
        {
            fGlWidget->renderConstraints = false;
            fGlWidget->renderScaleField = true;
        }
        else if(fGlWidget->renderScaleField)
        {
            fGlWidget->renderConstraints = false;
            fGlWidget->renderScaleField = false;
        }
        else
        {
            fGlWidget->renderConstraints = true;
            fGlWidget->renderScaleField = false;
        }
        fGlWidget->updateGL();
    }
    else if(e->key() == Qt::Key_M)
    {
        //fGlWidget->renderActualMesh = !fGlWidget->renderActualMesh;
        //fGlWidget->updateGL();
    }
    else if(e->key() == Qt::Key_Space)
    {
        //fGlWidget->startInterpolation();
    }
    else if(e->key() == Qt::Key_R)
    {
        //double theta = QInputDialog::getDouble(0,tr("Rotate 3D"), tr("Rotate by how many degrees?"));
        //mesh->rotate3DMeshAroundZBy(theta);
        //fGlWidget->updateGL();
    }
    else if(e->key() == Qt::Key_PageUp)
    {
        //fGlWidget->camz_interp = fGlWidget->camz_interp + 0.05;
        //fGlWidget->updateGL();
    }
    else if(e->key() == Qt::Key_PageDown)
    {
        //fGlWidget->camz_interp = fGlWidget->camz_interp - 0.05;
        //fGlWidget->updateGL();
    }
    else if(e->key() == Qt::Key_N)
    {
        //fGlWidget->renderNormals = !fGlWidget->renderNormals;
        //fGlWidget->updateGL();
    }
    else if(e->key() == Qt::Key_M)
    {
        fGlWidget->smoothZoom = !fGlWidget->smoothZoom;
    }
    else if(e->key() == Qt::Key_0)
    {
        fGlWidget->renderEnhanced = !fGlWidget->renderEnhanced;
        fGlWidget->updateGL();
    }
    else if(e->key() == Qt::Key_B)
    {
        //fGlWidget->removeBlackBorderBySize();
    }
}

void ParchmentFlattener::snapshot()
{
    int screenMult = QInputDialog::getInt(0,tr(""),tr("screen multiplier"));
    if(screenMult <= 0)
        return;
    fGlWidget->renderScreenToQImage(fGlWidget->window_width * screenMult, fGlWidget->window_height * screenMult);
}

void ParchmentFlattener::nextConstraint()
{
    if(ParchmentFlattener::inLineMode() || ParchmentFlattener::markupMode == ParchmentFlattener::EditLine)
    {
        mesh->selectedLineConstraintForward();
    }
    else if(ParchmentFlattener::inScaleMode() || ParchmentFlattener::markupMode == ParchmentFlattener::EditScale)
    {
        mesh->selectedScaleConstraintForward();
        int ssc = mesh->selectedScaleConstraint;
        if(ssc >= 0)
        {
            ScaleConstraint sc = mesh->scaleConstraints2[ssc];
            if(fGlWidget->renderOriginal)
            {
                fGlWidget->setTarget(mesh->getWrinkledSurfacePoint(sc.face,sc.u,sc.v));
            }
            else
            {
                fGlWidget->setTarget(mesh->getFlatSurfacePoint(sc.face,sc.u,sc.v));
            }
            fGlWidget->updateGL();
        }
    }
    fGlWidget->updateGL();
}

void ParchmentFlattener::prevConstraint()
{
    if(ParchmentFlattener::inLineMode() || ParchmentFlattener::markupMode == ParchmentFlattener::EditLine)
    {
        mesh->selectedLineConstraintBackward();
    }
    else if(ParchmentFlattener::inScaleMode() || ParchmentFlattener::markupMode == ParchmentFlattener::EditScale)
    {
        mesh->selectedScaleConstraintBackward();
        int ssc = mesh->selectedScaleConstraint;
        if(ssc >= 0)
        {
            ScaleConstraint sc = mesh->scaleConstraints2[ssc];
            if(fGlWidget->renderOriginal)
            {
                fGlWidget->setTarget(mesh->getWrinkledSurfacePoint(sc.face,sc.u,sc.v));
            }
            else
            {
                fGlWidget->setTarget(mesh->getFlatSurfacePoint(sc.face,sc.u,sc.v));
            }
            fGlWidget->updateGL();
        }
    }
    fGlWidget->updateGL();
}

void ParchmentFlattener::deleteConstraint()
{
    if(ParchmentFlattener::markupMode == ParchmentFlattener::EditLine)
    {
        mesh->deleteCurrentLineConstraint();
        setLineMode();
        fGlWidget->updateGL();
    }
    else if(ParchmentFlattener::markupMode == ParchmentFlattener::EditScale)
    {
        mesh->deleteCurrentScaleConstraint();
        int ssc = mesh->selectedScaleConstraint;
        if(ssc >= 0)
        {
            ScaleConstraint sc = mesh->scaleConstraints2[ssc];
            fGlWidget->setTarget(mesh->getFlatSurfacePoint(sc.face,sc.u,sc.v));
        }
        fGlWidget->updateGL();
    }
}

void ParchmentFlattener::toggleLineMode()
{
    if(ParchmentFlattener::inLineMode())
    {
        setNoneMode();
        mesh->floatingLineConstraint.clear();
    }
    else
    {
        setLineMode();
    }
    fGlWidget->updateGL();
}

void ParchmentFlattener::toggleScaleMode()
{
    if(ParchmentFlattener::inScaleMode())
    {
        setNoneMode();
        mesh->floatingScaleConstraint.clear();
    }
    else
    {
        setScaleMode();
    }
    fGlWidget->updateGL();
}

void ParchmentFlattener::toggleRenderCrossField()
{
    mesh->renderCrossField = !mesh->renderCrossField;
    fGlWidget->updateGL();
}

void ParchmentFlattener::saveConstraintFile()
{
    QString filename = QFileDialog::getSaveFileName(0,tr("Save Constraints"),ParchmentFlattener::meshDirectory,tr("Text files (*.txt)"));
    mesh->saveConstraints(filename);
}

void ParchmentFlattener::autosaveConstraintFile()
{
    QStringList nameFilter("autosave*");
    QDir directory(meshDirectory);
    QStringList autosaveFiles = directory.entryList(nameFilter, QDir::NoFilter, QDir::Time);
    int num_autosaves = autosaveFiles.count();
    int new_autosave_number = 0;
    QString autosave_filename = meshDirectory;

    if(num_autosaves > 0)
    {
        QString last_autosave_name = autosaveFiles.first();
        last_autosave_name.chop(4);
        QStringList tokens = last_autosave_name.split("_");
        int last_autosave_number = tokens[1].toInt();
        new_autosave_number = last_autosave_number + 1;
    }

    if(num_autosaves >= 5)
    {
        for(int k = 4; k < num_autosaves; ++k)
        {
            QString autosave_k = autosaveFiles[k];
            QString delete_filename = meshDirectory;
            delete_filename.append(autosave_k);
            QFile::remove(delete_filename);
        }
    }

    autosave_filename.append("autosave_").append(QString::number(new_autosave_number)).append(".txt");
    qDebug() << autosave_filename;
    mesh->saveConstraints(autosave_filename);
}

void ParchmentFlattener::loadConstraintFile()
{
    QString filename = QFileDialog::getOpenFileName(0,tr("Load Constraints"),ParchmentFlattener::meshDirectory,tr("Text files (*.txt)"));
    mesh->loadConstraints(filename);
}

void ParchmentFlattener::saveFlatMeshFile()
{
    QString filename = QFileDialog::getSaveFileName(0,tr("Save Constraints"),ParchmentFlattener::meshDirectory,tr("Obj files (*.obj)"));
    mesh->saveFlatMesh(filename);
}

void ParchmentFlattener::setRenderModeStandard()
{
    fGlWidget->renderStretch = false;
    fGlWidget->updateGL();
}

void ParchmentFlattener::setRenderModeStretch()
{
    fGlWidget->renderStretch = true;
    fGlWidget->updateGL();
}
